import markdown
from textwrap import dedent

from ford.md_admonition import AdmonitionExtension, FordMarkdownError

from bs4 import BeautifulSoup
import pytest


def convert(text: str) -> str:
    md = markdown.Markdown(extensions=[AdmonitionExtension()], output_format="html")
    return md.convert(dedent(text))


def not_title(tag):
    return tag.name == "p" and not tag.has_attr("class")


def test_basic():
    converted = convert("""
        @note note text
        """)

    soup = BeautifulSoup(converted, features="html.parser")
    assert len(soup) == 1
    assert sorted(soup.div["class"]) == ["alert", "alert-info"]
    assert soup.find(class_="h4").text == "Note"
    assert soup.div.find(not_title).text == "note text"


def test_uppercase():
    converted = convert("""
        @NOTE note text
        """)

    soup = BeautifulSoup(converted, features="html.parser")
    assert len(soup) == 1
    assert sorted(soup.div["class"]) == ["alert", "alert-info"]
    assert soup.find(class_="h4").text == "Note"
    assert soup.div.find(not_title).text == "note text"


def test_endnote():
    converted = convert("""
        @note note text @endnote
        """)

    soup = BeautifulSoup(converted, features="html.parser")
    assert len(soup) == 1
    assert sorted(soup.div["class"]) == ["alert", "alert-info"]
    assert soup.find(class_="h4").text == "Note"
    assert soup.div.find(not_title).text == "note text"


def test_paragraph():
    converted = convert("""
        @note note text
        some following text
        """)

    soup = BeautifulSoup(converted, features="html.parser")
    assert len(soup) == 1
    assert sorted(soup.div["class"]) == ["alert", "alert-info"]
    assert soup.find(class_="h4").text == "Note"
    assert soup.div.find(not_title).text == "note text\nsome following text"


def test_paragraph_end_line():
    converted = convert("""
        @note note text
        some following text""")

    soup = BeautifulSoup(converted, features="html.parser")
    assert len(soup) == 1
    assert sorted(soup.div["class"]) == ["alert", "alert-info"]
    assert soup.find(class_="h4").text == "Note"
    assert soup.div.find(not_title).text == "note text\nsome following text"


def test_explicit_end():
    converted = convert("""
        @note note text
        some blank lines


        before the end
        @endnote
        """)

    soup = BeautifulSoup(converted, features="html.parser")
    assert len(soup) == 1
    assert sorted(soup.div["class"]) == ["alert", "alert-info"]
    assert soup.find(class_="h4").text == "Note"
    assert soup.div.text.strip() == "Note\nnote text\nsome blank lines\nbefore the end"


def test_explicit_end_no_newline():
    converted = convert("""
        @note note text
        some blank lines


        before the end
        @endnote""")

    soup = BeautifulSoup(converted, features="html.parser")
    assert len(soup) == 1
    assert sorted(soup.div["class"]) == ["alert", "alert-info"]
    assert soup.find(class_="h4").text == "Note"
    assert soup.div.text.strip() == "Note\nnote text\nsome blank lines\nbefore the end"


def test_warning():
    converted = convert("""
        @warning note text
        """)

    soup = BeautifulSoup(converted, features="html.parser")
    assert len(soup) == 1
    assert sorted(soup.div["class"]) == ["alert", "alert-warning"]
    assert soup.find(class_="h4").text == "Warning"
    assert soup.div.find(not_title).text == "note text"


def test_in_list():
    converted = convert("""
        - item 1

            @note note text

        - item 2
        """)

    soup = BeautifulSoup(converted, features="html.parser")
    assert len(soup) == 1
    assert sorted(soup.div["class"]) == ["alert", "alert-info"]
    assert soup.find(class_="h4").text == "Note"
    assert soup.div.find(not_title).text == "note text"


def test_in_list_with_paragraph():
    converted = convert("""
        - item 1

            @note note text
            with a paragraph

            but not this one

        - item 2
        """)

    soup = BeautifulSoup(converted, features="html.parser")
    assert len(soup) == 1
    assert sorted(soup.div["class"]) == ["alert", "alert-info"]
    assert soup.find(class_="h4").text == "Note"
    assert soup.div.find(not_title).text == "note text\nwith a paragraph"


def test_in_list_with_paragraph_and_end():
    converted = convert("""
        - item 1

            @note note text
            with a paragraph

            and another one
            @endnote

        - item 2
        """)

    soup = BeautifulSoup(converted, features="html.parser")
    assert len(soup) == 1
    assert sorted(soup.div["class"]) == ["alert", "alert-info"]
    assert soup.find(class_="h4").text == "Note"
    all_text = "\n".join(p.text for p in soup.div.find_all(not_title))
    assert all_text == "note text\nwith a paragraph\nand another one"


def test_in_list_with_list():
    converted = convert("""
        - item 1

            @note note text

            1. first
            2. second

            @endnote

        - item 2
        """)

    soup = BeautifulSoup(converted, features="html.parser")
    assert len(soup) == 1
    assert sorted(soup.div["class"]) == ["alert", "alert-info"]
    assert soup.find(class_="h4").text == "Note"
    list_text = " ".join(li.text for li in soup.div.ol.find_all("li"))
    assert list_text == "first second"


# The following test doesn't currently work, but could be supported by
# keeping track of a stack of admonitions. PRs welcome!

# def test_in_list_with_nested_warning():
#     converted = convert(
#         """
#         - item 1

#             @note note text
#             with a paragraph

#                 @warning nested warning

#             unnested paragraph
#             @endnote

#         - item 2
#         """
#     )

#     soup = BeautifulSoup(converted, features="html.parser")
#     assert len(soup) == 1
#     assert sorted(soup.div["class"]) == ["alert", "alert-info"]
#     assert soup.find(class_="h4").text == "Note"

#     nested_box = soup.div.div.extract()
#     assert nested_box.find(class_="h4").text == "Warning"
#     assert nested_box.find(not_title).text == "nested warning"

#     all_text = "\n".join(p.text for p in soup.div.find_all(not_title))
#     assert all_text == "note text\nwith a paragraph\nunnested paragraph"


def test_midparagraph():
    converted = convert("""
        @Bug start text

        end text @endbug post text
        """)

    soup = BeautifulSoup(converted, features="html.parser")
    assert len(soup) == 3
    assert sorted(soup.div["class"]) == ["alert", "alert-danger"]
    assert soup.find(class_="h4").text == "Bug"
    all_text = "\n".join(p.text for p in soup.div.find_all(not_title))
    assert all_text == "start text\nend text"

    soup.div.extract()
    assert soup.text.strip() == "post text"


def test_title():
    converted = convert("""
        @warning some "title" text
        over multiple lines""")

    soup = BeautifulSoup(converted, features="html.parser")
    assert len(soup) == 1
    assert sorted(soup.div["class"]) == ["alert", "alert-warning"]
    assert soup.div.find(not_title).text == 'some "title" text\nover multiple lines'


def test_end_marker_without_start():
    with pytest.raises(FordMarkdownError):
        convert("@endnote")


def test_end_marker_doesnt_match_start():
    with pytest.raises(FordMarkdownError):
        convert("@bug\n@endnote")



from pathlib import Path
from ford.settings import ProjectSettings
from ford.pagetree import get_page_tree
from ford._markdown import MetaMarkdown

from textwrap import dedent
from locale import getpreferredencoding


def test_footnotes_on_one_page(tmp_path):
    """Check that footnotes only appear on the pages they are defined on,
    issue #327"""

    with open(tmp_path / "index.md", "w") as f:
        f.write("title: Index")

    with open(tmp_path / "a.md", "w") as f:
        f.write(dedent("""\
            title: Page A

            This has a footnote[^1] that should only appear on this page

            [^1]: This is the footnote on page A
            """))

    with open(tmp_path / "b.md", "w") as f:
        f.write(dedent("""\
            title: Page B

            This page should not have any footnotes
            """))

    md = MetaMarkdown()
    settings = ProjectSettings()
    result_dir = tmp_path / "result"
    nodes = get_page_tree(tmp_path, result_dir, tmp_path / "doc", md, settings)

    assert len(nodes.subpages) == 2
    assert "This is the footnote on page A" in nodes.subpages[0].contents
    assert "This is the footnote on page A" not in nodes.subpages[1].contents


def test_footnotes_on_one_page_parse_failure(tmp_path):
    """Check that footnotes only appear on the pages they are defined on,
    even if we fail to parse one of the files, issue #327"""

    with open(tmp_path / "index.md", "w") as f:
        f.write("title: Index")

    with open(tmp_path / "a.md", "w") as f:
        f.write(dedent("""\
            This page is missing a title and so expected not to be parsed

            This has a footnote[^1] that should only appear on this page

            [^1]: This is the footnote on page A
            """))

    with open(tmp_path / "b.md", "w") as f:
        f.write(dedent("""\
            title: Page B

            This page should not have any footnotes
            """))

    md = MetaMarkdown()
    settings = ProjectSettings()
    result_dir = tmp_path / "result"
    nodes = get_page_tree(tmp_path, result_dir, tmp_path / "doc", md, settings)

    assert len(nodes.subpages) == 1
    assert "This is the footnote on page A" not in nodes.subpages[0].contents


def test_non_utf8_encoding(tmp_path):
    """This is not really a full test for issue #518, as it only tests
    the lower part of the call-tree. A more thorough test would do
    something like encode a whole project in the non-default encoding.

    """
    # Try to get an encoding which is *not* the default for `open`
    encoding = "gbk" if getpreferredencoding().lower() == "utf-8" else "utf-8"

    with open(tmp_path / "index.md", "wb") as f:
        f.write(dedent("""\
        ---
        title: Specification
        ---

        @warning
        本文档是一个简易的规范文档，仅供参考
        """).encode(encoding))

    md = MetaMarkdown()
    settings = ProjectSettings()
    result_dir = tmp_path / "result"
    nodes = get_page_tree(
        tmp_path, result_dir, tmp_path / "doc", md, settings, encoding=encoding
    )

    assert "本文档是" in nodes.contents


def test_exclude_ford_dirs(tmp_path, capsys):
    """Test that we exclude some user-settable directories"""

    (tmp_path / "index.md").write_text("title: Index")

    subdir = tmp_path / "subdir"
    subdir.mkdir(parents=True)

    media_dir: Path = tmp_path / "media"
    media_dir.mkdir(parents=True)

    md = MetaMarkdown()
    settings = ProjectSettings(media_dir=media_dir)
    result_dir = tmp_path / "result"
    get_page_tree(tmp_path, result_dir, tmp_path / "doc", md, settings)

    captured = capsys.readouterr().out.replace("\n", " ").replace("  ", " ")
    assert str(media_dir) not in captured
    assert str(subdir) in captured



# -*- coding: utf-8 -*-
#
#  output.py
#  This file is part of FORD.
#
#  Copyright 2015 Christopher MacMackin <cmacmackin@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from dataclasses import asdict
import sys
import os
import shutil
import traceback
from itertools import chain
import pathlib
import time
from typing import List, Union, Callable, Type, Tuple
from warnings import simplefilter

from bs4 import BeautifulSoup, MarkupResemblesLocatorWarning
import jinja2

from ford.console import warn
from ford.sourceform import FortranBase
import ford.tipue_search
from ford.utils import ProgressBar
from ford.graphs import graphviz_installed, GraphManager
from ford.settings import ProjectSettings, EntitySettings

loc = pathlib.Path(__file__).parent
env = jinja2.Environment(
    trim_blocks=True,
    lstrip_blocks=True,
)
env.globals["path"] = os.path  # this lets us call path.* in templates

# Ignore bs4 warning about parsing strings that look like filenames
simplefilter("ignore", MarkupResemblesLocatorWarning)


def is_more_than_one(collection):
    return collection > 1


def meta(entity, item):
    """Get item from entity's meta dict, but give a more helpful
    error message if entity doesn't have a meta attribute

    Hopefully gives a better error than the common "RuntimeError: 'str
    object' has no attribute 'meta'"

    """
    if not hasattr(entity, "meta"):
        return jinja2.StrictUndefined(
            f"Unknown entity '{entity}': This likely means an error in parsing, "
            "please check that this file compiles with a Fortran compiler"
        )

    return getattr(entity.meta, item, None)


def relative_url(entity: Union[FortranBase, str], page_url: pathlib.Path) -> str:
    """Convert any links with absolute paths to the output directory
    to relative paths to ``page_url``
    """
    if isinstance(entity, str) and "/" not in entity or entity is None:
        return entity

    # Find first link in `entity` and get the path. If `entity`
    # doesn't have any links, it might be a URL itself that needs
    # fixing
    link_str = str(entity)
    link = BeautifulSoup(link_str, features="html.parser").a
    if link is not None:
        link_href = str(link["href"])
        if link_href.startswith("http"):
            # This is (almost certainly) an external link, so better
            # be correct already
            return entity
        link_path = str(pathlib.Path(link_href).resolve())
    else:
        link_path = link_str

    new_path = os.path.relpath(link_path, page_url.parent)
    return link_str.replace(link_path, new_path)


env.tests["more_than_one"] = is_more_than_one
env.filters["meta"] = meta
env.filters["relurl"] = relative_url

USER_WRITABLE_ONLY = 0o755


class Documentation:
    """
    Represents and handles the creation of the documentation files from
    a project.
    """

    def __init__(self, settings: ProjectSettings, proj_docs: str, project, pagetree):
        # This lets us use meta data anywhere within the template.
        # Also, in future for other template, we may not need to
        # pass the data obj.
        env.globals["projectData"] = asdict(settings)
        env.loader = jinja2.FileSystemLoader(
            settings.html_template_dir + [loc / "templates"]
        )

        self.project = project
        self.settings = settings
        # Jinja2's `if` statement counts `None` as truthy, so to avoid
        # lots of refactoring and messiness in the templates, just get
        # rid of None values
        self.data = {k: v for k, v in asdict(settings).items() if v is not None}
        self.data["pages"] = pagetree
        # Remove "project_url" so we can set it as a relative path for
        # each page individually and not have it clobbered by this dict
        del self.data["project_url"]
        self.lists: List[ListPage] = []
        self.docs = []
        self.njobs = settings.parallel
        self.parallel = self.njobs > 0

        self.index = IndexPage(self.data, project, proj_docs)
        self.search = SearchPage(self.data, project)
        if not graphviz_installed and settings.graph:
            warn("Will not be able to generate graphs. Graphviz not installed.")
        if settings.relative:
            graphparent = "../"
        else:
            graphparent = ""

        print("  Creating HTML documentation... ", end="")
        html_time_start = time.time()
        try:
            PageFactory = Union[Type, Callable]

            # Create individual entity pages
            entity_list_page_map: List[Tuple[List, PageFactory]] = [
                (project.types, TypePage),
                (project.absinterfaces, AbsIntPage),
                (project.procedures, ProcPage),
                (project.submodprocedures, ProcPage),
                (project.modules, ModulePage),
                (project.submodules, ModulePage),
                (project.programs, ProgPage),
                (project.blockdata, BlockPage),
                (project.namelists, NamelistPage),
            ]
            if settings.incl_src:
                entity_list_page_map.append((project.allfiles, FilePage))

            for entity_list, page_class in entity_list_page_map:
                for item in entity_list:
                    self.docs.append(page_class(self.data, project, item))

            # Create lists of each entity type
            if len(project.procedures) > 0:
                self.lists.append(ProcList(self.data, project))
            if settings.incl_src and (
                len(project.files) + len(project.extra_files) > 1
            ):
                self.lists.append(FileList(self.data, project))
            if len(project.modules) + len(project.submodules) > 0:
                self.lists.append(ModList(self.data, project))
            if len(project.programs) > 1:
                self.lists.append(ProgList(self.data, project))
            if len(project.types) > 0:
                self.lists.append(TypeList(self.data, project))
            if len(project.absinterfaces) > 0:
                self.lists.append(AbsIntList(self.data, project))
            if len(project.blockdata) > 1:
                self.lists.append(BlockList(self.data, project))
            if project.namelists:
                self.lists.append(NamelistList(self.data, project))

            # Create static pages
            self.pagetree = [
                PagetreePage(self.data, project, item) for item in (pagetree or [])
            ]
        except Exception:
            if settings.dbg:
                traceback.print_exc()
                sys.exit("Error encountered.")
            else:
                sys.exit('Error encountered. Run with "--debug" flag for traceback.')

        html_time_end = time.time()
        print(f"done in {html_time_end - html_time_start:5.3f}s")

        self.graphs = GraphManager(
            self.data.get("graph_dir", ""),
            graphparent,
            settings.coloured_edges,
            settings.show_proc_parent,
            save_graphs=bool(self.data.get("graph_dir", False)),
        )

        if graphviz_installed and settings.graph:
            for entity_list in [
                project.types,
                project.procedures,
                project.submodprocedures,
                project.modules,
                project.submodules,
                project.programs,
                project.files,
                project.blockdata,
            ]:
                for item in entity_list:
                    self.graphs.register(item)

            self.graphs.graph_all()
            project.callgraph = self.graphs.callgraph
            project.typegraph = self.graphs.typegraph
            project.usegraph = self.graphs.usegraph
            project.filegraph = self.graphs.filegraph
        else:
            project.callgraph = ""
            project.typegraph = ""
            project.usegraph = ""
            project.filegraph = ""

        if settings.search:
            url = "" if settings.relative else settings.project_url
            self.tipue = ford.tipue_search.Tipue_Search_JSON_Generator(
                settings.output_dir, url
            )
            self.tipue.create_node(
                self.index.html, "index.html", EntitySettings(category="home")
            )
            jobs = len(self.docs) + len(self.pagetree)
            for page in (
                bar := ProgressBar(
                    "Creating search index", chain(self.docs, self.pagetree), total=jobs
                )
            ):
                bar.set_current(page.loc)
                self.tipue.create_node(page.html, page.loc, page.meta)

    def writeout(self) -> None:
        out_dir: pathlib.Path = self.data["output_dir"]
        # Remove any existing file/directory. This avoids errors coming from
        # `shutils.copytree` for Python < 3.8, where we can't explicitly ignore them
        if out_dir.is_file():
            out_dir.unlink()
        else:
            shutil.rmtree(out_dir, ignore_errors=True)

        try:
            out_dir.mkdir(USER_WRITABLE_ONLY, parents=True)
        except Exception as e:
            print(f"Error: Could not create output directory. {e.args[0]}")

        for directory in [
            "lists",
            "sourcefile",
            "type",
            "proc",
            "interface",
            "module",
            "program",
            "src",
            "blockdata",
            "namelist",
        ]:
            (out_dir / directory).mkdir(USER_WRITABLE_ONLY)

        for directory in ["css", "js", "webfonts"]:
            copytree(loc / directory, out_dir / directory)

        if self.data["graph"]:
            self.graphs.output_graphs(self.njobs)
        if self.data["search"]:
            copytree(loc / "search", out_dir / "search")
            self.tipue.print_output()

        try:
            copytree(self.data["media_dir"], out_dir / "media")
        except OSError as e:
            warn(f"error copying media directory {self.data['media_dir']}, {e}")
        except KeyError:
            pass

        if "css" in self.data:
            shutil.copy(self.data["css"], out_dir / "css" / "user.css")

        shutil.copy(self.data["favicon"], out_dir / "favicon.png")

        if self.data["incl_src"]:
            for src in self.project.allfiles:
                shutil.copy(src.path, out_dir / "src" / src.name)

        if "mathjax_config" in self.data:
            mathjax_path = out_dir / "js" / "MathJax-config"
            mathjax_path.mkdir(parents=True, exist_ok=True)
            shutil.copy(
                self.data["mathjax_config"],
                mathjax_path / os.path.basename(self.data["mathjax_config"]),
            )

        items = list(
            chain(self.docs, self.lists, self.pagetree, [self.index, self.search])
        )
        for page in (bar := ProgressBar("Writing files", items)):
            bar.set_current(os.path.relpath(page.outfile))
            page.writeout()

        print(f"\nBrowse the generated documentation: file://{out_dir}/index.html")


class BasePage:
    """
    Abstract class for representation of pages in the documentation.

      data
        Dictionary containing project information (to be used when rendering)
      proj
        FortranProject object
      obj
        The object/item in the code which this page is documenting
    """

    def __init__(self, data, proj, obj=None):
        self.data = data
        self.proj = proj
        self.obj = obj
        self.meta = getattr(obj, "meta", EntitySettings())
        self.out_dir = self.data["output_dir"]
        self.page_dir = self.out_dir / "page"
        self.relative = data["relative"]
        self.project_url = (
            os.path.relpath(proj.settings.project_url, self.outfile.parent)
            if self.relative
            else proj.settings.project_url
        )

    @property
    def html(self) -> str:
        """Wrapper for only doing the rendering on request (drastically reduces memory)"""
        try:
            return self.render(self.data, self.proj, self.obj)
        except Exception as e:
            raise RuntimeError(
                f"Error rendering '{self.outfile.name}':\n"
                f'  File "{self.obj.filename}": {self.obj.obj} "{self.obj.name}"\n'
                f"    {e}"
            )

    @property
    def outfile(self) -> pathlib.Path:
        raise NotImplementedError()

    def writeout(self) -> None:
        self.outfile.write_bytes(self.html.encode("utf8"))

    @property
    def template_path(self) -> str:
        """Path to page template file (relative to 'templates/')"""
        raise NotImplementedError()

    @property
    def template(self):
        """Jinja template loaded from `template_path` with globals set"""
        return env.get_template(
            self.template_path,
            globals=dict(page_url=self.outfile, project_url=self.project_url),
        )

    def render(self, data, project, object):
        """
        Get the HTML for the page. This method must be overridden. Arguments
        are proj_data, project object, and item in the code which the
        page documents.
        """
        raise NotImplementedError("Should not instantiate BasePage type")


class ListTopPage(BasePage):
    @property
    def outfile(self):
        return self.out_dir / self.template_path

    def render(self, data, project, object):
        return self.template.render(data, project=project, proj_docs=object)


class IndexPage(ListTopPage):
    template_path = "index.html"


class SearchPage(ListTopPage):
    template_path = "search.html"


class ListPage(BasePage):
    @property
    def out_page(self):
        raise NotImplementedError("ListPage subclass missing 'out_page' property")

    @property
    def outfile(self):
        return self.out_dir / "lists" / self.out_page

    def render(self, data, project, object):
        return self.template.render(data, project=project)


class ProcList(ListPage):
    out_page = "procedures.html"
    template_path = "proc_list.html"


class FileList(ListPage):
    out_page = "files.html"
    template_path = "file_list.html"


class ModList(ListPage):
    out_page = "modules.html"
    template_path = "mod_list.html"


class ProgList(ListPage):
    out_page = "programs.html"
    template_path = "prog_list.html"


class TypeList(ListPage):
    out_page = "types.html"
    template_path = "types_list.html"


class AbsIntList(ListPage):
    out_page = "absint.html"
    template_path = "absint_list.html"


class BlockList(ListPage):
    out_page = "blockdata.html"
    template_path = "block_list.html"


class NamelistList(ListPage):
    out_page = "namelists.html"
    template_path = "namelist_list.html"


class DocPage(BasePage):
    """
    Abstract class to be inherited by all pages for items in the code.
    """

    @property
    def payload_key(self):
        raise NotImplementedError("DocPage subclass missing 'payload_key'")

    @property
    def object_page(self):
        return self.obj.ident + ".html"

    @property
    def loc(self):
        return pathlib.Path(self.obj.get_dir()) / self.object_page

    @property
    def outfile(self):
        return self.out_dir / self.obj.get_dir() / self.object_page

    def render(self, data, project, object):
        try:
            return self.template.render(
                data, project=project, **{self.payload_key: object}
            )
        except jinja2.exceptions.TemplateError:
            print(f"Error rendering page '{self.outfile}'")
            raise


class FilePage(DocPage):
    template_path = "file_page.html"
    payload_key = "src"


class TypePage(DocPage):
    template_path = "type_page.html"
    payload_key = "dtype"


class AbsIntPage(DocPage):
    template_path = "nongenint_page.html"
    payload_key = "interface"


class ModulePage(DocPage):
    template_path = "mod_page.html"
    payload_key = "module"


class ProgPage(DocPage):
    template_path = "prog_page.html"
    payload_key = "program"


class BlockPage(DocPage):
    template_path = "block_page.html"
    payload_key = "blockdat"


class ProcedurePage(DocPage):
    template_path = "proc_page.html"
    payload_key = "procedure"


class GenericInterfacePage(DocPage):
    template_path = "genint_page.html"
    payload_key = "interface"


class InterfacePage(DocPage):
    template_path = "nongenint_page.html"
    payload_key = "interface"


class NamelistPage(DocPage):
    template_path = "namelist_page.html"
    payload_key = "namelist"


def ProcPage(data, proj, obj):
    """Factory function for creating procedure or interface pages"""
    if obj.obj == "proc":
        return ProcedurePage(data, proj, obj)
    if obj.generic:
        return GenericInterfacePage(data, proj, obj)
    return InterfacePage(data, proj, obj)


class PagetreePage(BasePage):
    template_path = "info_page.html"

    @property
    def loc(self):
        return pathlib.Path("page") / self.obj.path

    @property
    def outfile(self):
        return self.page_dir / self.obj.path

    def render(self, data, project, object):
        return self.template.render(
            data, page=object, project=project, topnode=object.topnode
        )

    def writeout(self):
        if self.obj.filename.stem == "index":
            (self.page_dir / self.obj.location).mkdir(USER_WRITABLE_ONLY, exist_ok=True)
        super(PagetreePage, self).writeout()

        from_path = self.data["page_dir"] / self.obj.location
        to_path = self.page_dir / self.obj.location

        for item in self.obj.copy_subdir:
            item_path = from_path / item
            try:
                copytree(item_path, to_path / item)
            except Exception as e:
                warn(f"could not copy directory '{item_path}'. Error: {e.args[0]}")

        for item in self.obj.files:
            item_path = from_path / item
            try:
                shutil.copy(item_path, to_path)
            except Exception as e:
                warn(f"could not copy file '{item_path}'. Error: {e.args[0]}")


def copytree(src: pathlib.Path, dst: pathlib.Path) -> None:
    """Wrapper around `shutil.copytree` that:
    a) doesn't try to set xattrs; and
    b) ensures modification time is time of current FORD run
    """
    shutil.copytree(src, dst, copy_function=shutil.copy)
    for file in dst.rglob("*"):
        file.touch()



import pytest
from textwrap import dedent
import ford
from ford.utils import meta_preprocessor


@pytest.mark.parametrize("string", ["true", "True", "TRUE", "tRuE"])
def test_str_to_bool_true(string):
    assert ford.utils.str_to_bool(string)


@pytest.mark.parametrize("string", ["false", "False", "FALSE", "fAlSe"])
def test_str_to_bool_false(string):
    assert not ford.utils.str_to_bool(string)


def test_str_to_bool_already_bool():
    assert ford.utils.str_to_bool(True)
    assert not ford.utils.str_to_bool(False)


@pytest.mark.parametrize(
    ("string", "level", "expected"),
    [
        ("abcdefghi", 0, ["abcdefghi"]),
        ("abc(def)ghi", 1, ["(def)"]),
        ("abc(def)ghi", 0, ["abc()ghi"]),
        ("(abc)def(ghi)", 1, ["(abc)", "(ghi)"]),
        ("(a(b)c)def(gh(i))", 1, ["(a()c)", "(gh())"]),
    ],
)
def test_strip_paren(string, level, expected):
    assert ford.utils.strip_paren(string, retlevel=level) == expected


def test_meta_preprocessor():
    text = dedent("""\
    key1: value1
    key2: value2
          value2a
    key3: value3

    no more metadata""")

    meta, doc = meta_preprocessor(text)

    assert doc == ["no more metadata"]
    assert meta == {
        "key1": ["value1"],
        "key2": ["value2", "value2a"],
        "key3": ["value3"],
    }



from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("ford")
except PackageNotFoundError:
    from setuptools_scm import get_version  # type: ignore[import]

    __version__ = get_version(root="..", relative_to=__file__)



"""
Admonition Preprocessor
=======================

Markdown preprocessor for dealing with FORD style admonitions. See
:ref:`sec-note-boxes` for details on using these in Ford docs.

A preprocessor, :py:class:`AdmonitionPreprocessor`, converts Ford style
``@note`` into something that the existing `markdown admonition
extension`_ can handle. This is mostly a matter of making sure the
note body is indented correctly. The conversion to HTML is done with a
customised processor, :py:class:`FordAdmonitionProcessor`

.. _markdown admonition extension:
   https://python-markdown.github.io/extensions/admonition/

"""

import re
from dataclasses import dataclass
from typing import ClassVar, List
from textwrap import indent

from markdown.extensions import Extension
from markdown.preprocessors import Preprocessor
from markdown.extensions.admonition import AdmonitionProcessor

ADMONITION_TYPE = {
    "note": "info",
    "warning": "warning",
    "todo": "success",
    "bug": "danger",
    "history": "history",
}
"""Mapping of Ford note types to markdown's admonition types"""


class AdmonitionExtension(Extension):
    """Admonition extension for Python-Markdown."""

    def extendMarkdown(self, md):
        """Add Admonition to Markdown instance."""
        md.registerExtension(self)
        md.parser.blockprocessors.deregister("admonition", strict=False)
        md.preprocessors.register(AdmonitionPreprocessor(md), "admonition-pre", 100)
        md.parser.blockprocessors.register(
            FordAdmonitionProcessor(md.parser), "admonition", 100
        )


class FordMarkdownError(RuntimeError):
    """Format an error when processing markdown, giving some context"""

    def __init__(
        self,
        message: str,
        line_number: int,
        lines: List[str],
        start: int,
        end: int,
        context: int = 4,
    ):
        line_start = line_number - context
        line_end = line_number + context
        line_context = lines[line_start:line_end]
        num_len = len(f"{line_end}")
        text_with_line_numbers = [
            f"{line_start + n + 1:{num_len}d}: {line}"
            for n, line in enumerate(line_context)
        ]
        marker = f"{' ' * (num_len + 2 + start)}{'^' * (end - start)}"
        text_with_line_numbers.insert(context + 1, marker)
        text = indent("\n".join(text_with_line_numbers), "    ")
        super().__init__(f"{message}:\n\n{text}")


class FordAdmonitionProcessor(AdmonitionProcessor):
    """Customised version of the `Python markdown`_ extension.

    Uses our CSS class names for each specific note type.

    .. _Python markdown:
       https://python-markdown.github.io/extensions/admonition/
    """

    CLASSNAME = "alert"
    CLASSNAME_TITLE = "alert-title h4"
    RE = re.compile(r"""(?:^|\n)@note ?(?P<klass>[\w\-]+) *(?:\n|$)""")

    def get_class_and_title(self, match):
        """Get the CSS class and title for this admonition

        Title is the note class, while the CSS class is looked up
        in the list of note types (`ADMONITION_TYPE`)

        """
        klass = match["klass"].lower()
        return f"alert-{ADMONITION_TYPE[klass]}", klass.capitalize()


class AdmonitionPreprocessor(Preprocessor):
    """Markdown preprocessor for dealing with FORD style admonitions.

    This preprocessor converts the FORD syntax for admonitions to
    the markdown admonition syntax.

    A FORD admonition starts with ``@<type>``, where ``<type>`` is one of:
    ``note``, ``warning``, ``todo``, ``bug``, or ``history``.
    An admonition ends at (in this order of preference):

    1. ``@end<type>``, where ``<type>`` must match the start marker
    2. an empty line
    3. a new note (``@<type>``)
    4. the end of the documentation lines

    The admonitions are converted to the markdown syntax, i.e. ``@note Note``,
    followed by an indented block. Possible end markers are removed, as well
    as empty lines if they mark the end of an admonition.
    """

    INDENT_SIZE: ClassVar[int] = 4
    INDENT: ClassVar[str] = " " * INDENT_SIZE
    ADMONITION_RE: ClassVar[re.Pattern] = re.compile(
        rf"""(?P<indent>\s*)
        @(?P<type>{"|".join(ADMONITION_TYPE.keys())})
        (?P<posttxt>.*)
        """,
        re.IGNORECASE | re.VERBOSE,
    )
    END_RE: ClassVar[re.Pattern] = re.compile(
        rf"""\s*@end(?P<type>{"|".join(ADMONITION_TYPE.keys())})
        \s*(?P<posttxt>.*)?""",
        re.IGNORECASE | re.VERBOSE,
    )
    admonitions: List["Admonition"] = []

    @dataclass
    class Admonition:
        """A single admonition block in the text."""

        #: Admonition type (note, bug, and so on)
        type: str
        #: Line index of start marker
        start_idx: int
        #: Line index where admonition ends
        end_idx: int = -1

    def run(self, lines: List[str]) -> List[str]:
        admonitions = self._find_admonitions(lines)
        return self._process_admonitions(admonitions, lines)

    def _find_admonitions(self, lines: List[str]) -> List[Admonition]:
        """Scans the lines to search for admonitions."""
        admonitions = []
        current_admonition = None

        for idx, line in enumerate(lines):
            if match := self.ADMONITION_RE.search(line):
                if current_admonition:
                    if current_admonition.end_idx == -1:
                        current_admonition.end_idx = idx
                    admonitions.append(current_admonition)
                current_admonition = self.Admonition(type=match["type"], start_idx=idx)

            if end := self.END_RE.search(line):
                if not current_admonition:
                    raise FordMarkdownError(
                        "Note end marker found without start marker",
                        idx,
                        lines,
                        end.start(),
                        end.end(),
                    )

                if end["type"].lower() != current_admonition.type.lower():
                    raise FordMarkdownError(
                        "Type of start and end marker don't match",
                        idx,
                        lines,
                        end.start(),
                        end.end(),
                    )

                current_admonition.end_idx = idx
                admonitions.append(current_admonition)
                current_admonition = None

            if current_admonition is None:
                continue

            if line == "" and current_admonition.end_idx == -1:
                # empty line encountered while in an admonition. We set end_line but don't
                # move it to the list yet since an end marker (@end...) may follow
                # later.
                current_admonition.end_idx = idx

        if current_admonition:
            # We reached the last line and the last admonition wasn't moved to the list yet.
            if current_admonition.end_idx == -1:
                current_admonition.end_idx = idx
            admonitions.append(current_admonition)

        return admonitions

    def _process_admonitions(
        self, admonitions: List[Admonition], lines: List[str]
    ) -> List[str]:
        """Processes the admonitions to convert the lines to the markdown syntax."""

        # We handle the admonitions in the reverse order since
        # we may be deleting lines.
        for admonition in admonitions[::-1]:
            # last line--deal with possible text before or after end marker
            idx = admonition.end_idx
            if end := self.END_RE.search(lines[idx]):
                # Shove anything after the @end into a new paragraph
                if end["posttxt"]:
                    lines.insert(idx + 1, "")
                    lines.insert(idx + 2, end["posttxt"])

                # Remove the @end and possibly the line too if it ends up blank
                lines[idx] = self.END_RE.sub("", lines[idx])
                if lines[idx].strip() == "":
                    del lines[idx]
                else:
                    # New ending is now next line
                    admonition.end_idx += 1

            # Indent any intermediate lines
            end_line = min(len(lines), admonition.end_idx + 1)
            for idx in range(admonition.start_idx + 1, end_line):
                if lines[idx] != "":
                    lines[idx] = self.INDENT + lines[idx]

            idx = admonition.start_idx
            if (match := self.ADMONITION_RE.search(lines[idx])) is None:
                # Something has gone seriously wrong!
                raise FordMarkdownError("Missing start of @note", idx, lines, 0, -1)

            lines[idx] = f"{match['indent']}@note {admonition.type.capitalize()}"
            if posttxt := match["posttxt"]:
                lines.insert(idx + 1, self.INDENT + match["indent"] + posttxt)

        return lines



from __future__ import annotations

import json
import pathlib
import re
from urllib.error import URLError
from urllib.request import urlopen
from urllib.parse import urljoin

from ford.sourceform import (
    FortranBase,
    FortranType,
    ExternalModule,
    ExternalFunction,
    ExternalSubroutine,
    ExternalInterface,
    ExternalType,
    ExternalVariable,
    ExternalBoundProcedure,
)
from ford.version import __version__

# attributes of a module object needed for further processing
ATTRIBUTES = [
    "pub_procs",
    "pub_absints",
    "pub_types",
    "pub_vars",
    "functions",
    "subroutines",
    "interfaces",
    "absinterfaces",
    "types",
    "variables",
    "boundprocs",
    "vartype",
    "permission",
    "deferred",
    "generic",
    "attribs",
]

# Mapping between entity name and its type
ENTITIES = {
    "module": ExternalModule,
    "interface": ExternalInterface,
    "type": ExternalType,
    "variable": ExternalVariable,
    "function": ExternalFunction,
    "subroutine": ExternalSubroutine,
    "boundprocedure": ExternalBoundProcedure,
}

METADATA_NAME = "ford-metadata"


def obj2dict(intObj):
    """
    Converts an object to a dictionary.
    """
    if hasattr(intObj, "external_url"):
        return None
    if isinstance(intObj, str):
        return intObj
    extDict = {
        "name": intObj.name,
        "external_url": f"./{intObj.get_url()}",
        "obj": intObj.obj,
    }
    if hasattr(intObj, "proctype"):
        extDict["proctype"] = intObj.proctype
    for attrib in ATTRIBUTES:
        if not hasattr(intObj, attrib):
            continue

        attribute = getattr(intObj, attrib)

        if isinstance(attribute, list):
            extDict[attrib] = [obj2dict(item) for item in attribute]
        elif isinstance(attribute, dict):
            extDict[attrib] = {key: obj2dict(val) for key, val in attribute.items()}
        else:
            extDict[attrib] = str(attribute)
    return extDict


def modules_from_local(url: pathlib.Path):
    """
    Get module information from an external project but on the
    local file system.
    """

    return json.loads((url / "modules.json").read_text(encoding="utf-8"))


def dict2obj(project, extDict, url, parent=None, remote: bool = False) -> FortranBase:
    """
    Converts a dictionary to an object and immediately adds it to the project
    """
    if isinstance(extDict, str):
        return extDict
    name = extDict["name"]
    if extDict["external_url"]:
        extDict["external_url"] = extDict["external_url"].split("/", 1)[-1]
        if remote:
            external_url = urljoin(url, extDict["external_url"])
        else:
            external_url = url / extDict["external_url"]
    else:
        external_url = extDict["external_url"]

    # Look up what type of entity this is
    obj_type = extDict.get("proctype", extDict["obj"]).lower()
    # Construct the entity
    extObj = ENTITIES[obj_type](name, external_url, parent)
    # Now add it to the correct project list
    project_list = getattr(project, extObj._project_list)
    project_list.append(extObj)

    if type(extObj) is ExternalInterface:
        extObj.proctype = extDict["proctype"]

    for key in ATTRIBUTES:
        if key not in extDict:
            continue
        if isinstance(extDict[key], list):
            tmpLs = [
                dict2obj(project, item, url, extObj, remote)
                for item in extDict[key]
                if item
            ]
            setattr(extObj, key, tmpLs)
        elif isinstance(extDict[key], dict):
            tmpDict = {
                key2: dict2obj(project, item, url, extObj, remote)
                for key2, item in extDict[key].items()
                if item
            }
            setattr(extObj, key, tmpDict)
        else:
            setattr(extObj, key, extDict[key])
    return extObj


def dump_modules(project, path="."):
    """Dump modules to JSON file"""

    modules = [obj2dict(module) for module in project.modules]
    data = {
        METADATA_NAME: {
            "version": __version__,
        },
        "modules": modules,
    }
    (pathlib.Path(path) / "modules.json").write_text(json.dumps(data))


def load_external_modules(project):
    """Load external modules from JSON file into an existing project"""

    # get the external modules from the external URLs
    for url in project.external.values():
        remote = re.match("https?://", url)
        try:
            if remote:
                # Ensure the URL ends with '/' to have urljoin work as
                # intentend.
                if url[-1] != "/":
                    url = url + "/"
                extModules = json.loads(
                    urlopen(urljoin(url, "modules.json")).read().decode("utf8")
                )
            else:
                if not pathlib.Path(url).is_absolute():
                    url = project.settings.directory.joinpath(url).resolve()
                extModules = modules_from_local(url)
        except (URLError, json.JSONDecodeError) as error:
            extModules = {}
            print(f"Could not open external URL '{url}', reason: {error}")

        if METADATA_NAME in extModules:
            # TODO: Check version compatibility
            extModules = extModules["modules"]

        # convert modules defined in the JSON database to module objects
        for extModule in extModules:
            dict2obj(project, extModule, url, remote=bool(remote))



from dataclasses import asdict, dataclass, field
from datetime import date
from itertools import combinations
from pathlib import Path
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)
from markdown_include.include import (
    INC_SYNTAX as MD_INCLUDE_RE,
    MarkdownInclude,
    IncludePreprocessor,
)

from ford._typing import PathLike
from ford.console import warn
from ford.utils import meta_preprocessor, normalise_path, str_to_bool

try:
    import tomllib  # type: ignore[import]
except ModuleNotFoundError:
    import tomli as tomllib  # type: ignore[no-redef]


FAVICON_PATH = Path("favicon.png")

INTRINSIC_MODS = {
    "iso_fortran_env": "http://fortranwiki.org/fortran/show/iso_fortran_env",
    "iso_c_binding": "http://fortranwiki.org/fortran/show/iso_c_binding",
    "ieee_arithmetic": "http://fortranwiki.org/fortran/show/ieee_arithmetic",
    "ieee_exceptions": "http://fortranwiki.org/fortran/show/IEEE+arithmetic",
    "ieee_features": "http://fortranwiki.org/fortran/show/IEEE+arithmetic",
    "openacc": "https://www.openacc.org/sites/default/files/inline-images/Specification/OpenACC.3.0.pdf#page=85",
    "omp_lib": "https://www.openmp.org/spec-html/5.1/openmpch3.html#x156-1890003",
    "mpi": "http://www.mpi-forum.org/docs/mpi-3.1/mpi31-report/node410.htm",
    "mpi_f08": "http://www.mpi-forum.org/docs/mpi-3.1/mpi31-report/node409.htm",
}

# Mapping from key to separator for settings that are dicts. Required
# due to the legacy format
OPTION_SEPARATORS = {
    "alias": "=",
    "external": "=",
    "extra_mods": ":",
    "extra_vartypes": ":",
}


def default_cpus() -> int:
    try:
        import multiprocessing

        return multiprocessing.cpu_count()
    except (ImportError, NotImplementedError):
        return 0


def is_same_type(type_in: Type, tp: Type) -> bool:
    """Returns True if ``type_in`` is the same type as either ``tp``
    or ``Optional[tp]``"""
    return (type_in == tp) or is_optional_type(type_in, tp)


def is_optional_type(tp: Type, sub_tp: Type) -> bool:
    """Returns True if ``tp`` is ``Optional[sub_tp]``"""
    if get_origin(tp) is not Union:
        return False

    return any(tp == sub_tp for tp in get_args(tp))


def convert_to_bool(name: str, option: List[str]) -> bool:
    """Convert value 'option' to a bool, with a nice error message on
    failure. Expects a list from the markdown meta-data extension"""
    if isinstance(option, bool):
        return option

    if len(option) > 1:
        raise ValueError(
            f"Could not convert option '{name}' to bool: expected a single value but got a list ({option})"
        )
    try:
        return str_to_bool(option[0])
    except ValueError:
        raise ValueError(
            f"Could not convert option '{name}' to bool: expected 'true'/'false', got: {option[0]}"
        )


@dataclass
class ExtraFileType:
    extension: str
    comment: str
    lexer: Optional[str] = None

    @classmethod
    def from_string(cls, string: str):
        parts = string.split()
        if not (2 <= len(parts) <= 3):
            raise ValueError(
                f"Unexpected format for 'extra_filetype': expected 'extension comment [lexer]', got {string!r}"
            )

        file_type = cls(parts[0], parts[1])
        if len(parts) == 3:
            file_type.lexer = parts[2]

        return file_type


@dataclass
class ProjectSettings:
    alias: Dict[str, str] = field(default_factory=dict)
    author: Optional[str] = None
    author_description: Optional[str] = None
    author_pic: Optional[str] = None
    bitbucket: Optional[str] = None
    coloured_edges: bool = False
    copy_subdir: List[Path] = field(default_factory=list)
    creation_date: str = "%Y-%m-%dT%H:%M:%S.%f%z"
    css: Optional[Path] = None
    dbg: bool = True
    directory: Path = Path.cwd()
    display: List[str] = field(default_factory=lambda: ["public", "protected"])
    doc_license: str = ""
    docmark: str = "!"
    doxygen: bool = True
    docmark_alt: str = "*"
    email: Optional[str] = None
    encoding: str = "utf-8"
    exclude: List[str] = field(default_factory=list)
    exclude_dir: List[Path] = field(default_factory=list)
    extensions: List[str] = field(
        default_factory=lambda: ["f90", "f95", "f03", "f08", "f15"]
    )
    external: Dict[str, str] = field(default_factory=dict)
    externalize: bool = True
    extra_filetypes: Dict[str, ExtraFileType] = field(default_factory=dict)
    extra_mods: Dict[str, str] = field(default_factory=dict)
    extra_vartypes: list = field(default_factory=list)
    facebook: Optional[str] = None
    favicon: Path = FAVICON_PATH
    fixed_extensions: List[str] = field(
        default_factory=lambda: ["f", "for", "F", "FOR"]
    )
    fixed_length_limit: bool = True
    force: bool = False
    fpp_extensions: List[str] = field(
        default_factory=lambda: ["F90", "F95", "F03", "F08", "F15", "F", "FOR"]
    )
    github: Optional[str] = None
    gitlab: Optional[str] = None
    gitter_sidecar: Optional[str] = None
    google_plus: Optional[str] = None
    graph: bool = False
    graph_dir: Optional[Path] = None
    graph_maxdepth: int = 10000
    graph_maxnodes: int = 1000000000
    hide_undoc: bool = False
    html_template_dir: List[Path] = field(default_factory=list)
    incl_src: bool = True
    include: List[Path] = field(default_factory=list)
    license: str = ""
    linkedin: Optional[str] = None
    lower: bool = False
    macro: List[str] = field(default_factory=list)
    mathjax_config: Optional[Path] = None
    max_frontpage_items: int = 10
    md_base_dir: Path = Path(".")
    md_extensions: List[str] = field(default_factory=list)
    media_dir: Optional[Path] = None
    output_dir: Path = Path("./doc")
    page_dir: Optional[Path] = None
    parallel: int = default_cpus()
    predocmark: str = ">"
    predocmark_alt: str = "|"
    preprocess: bool = True
    preprocessor: str = "pcpp -D__GFORTRAN__ --passthru-comments"
    print_creation_date: bool = False
    privacy_policy_url: Optional[str] = None
    proc_internals: bool = False
    project: str = "Fortran Program"
    project_bitbucket: Optional[str] = None
    project_download: Optional[str] = None
    project_github: Optional[str] = None
    project_gitlab: Optional[str] = None
    project_sourceforge: Optional[str] = None
    project_url: str = ""
    project_website: Optional[str] = None
    quiet: bool = False
    relative: bool = field(init=False)
    revision: Optional[str] = None
    search: bool = True
    show_proc_parent: bool = False
    sort: str = "src"
    source: bool = False
    src_dir: List[Path] = field(default_factory=lambda: [Path("./src")])
    summary: Optional[str] = None
    terms_of_service_url: Optional[str] = None
    twitter: Optional[str] = None
    version: Optional[str] = None
    warn: bool = False
    website: Optional[str] = None
    year: str = str(date.today().year)

    def __post_init__(self):
        self.relative = self.project_url == ""

        field_types = get_type_hints(ProjectSettings)

        for key, value in asdict(self).items():
            default_type = field_types[key]

            if is_same_type(default_type, type(value)):
                continue

            if get_origin(default_type) is list and not isinstance(value, list):
                setattr(self, key, [value])

        for fixed_extension in self.fixed_extensions:
            if fixed_extension in self.extensions:
                raise ValueError(
                    f"Fixed-form extension '{fixed_extension}' also appears in free-form extension list (`extensions = {self.extensions}`)"
                )

        for mod in self.extra_mods:
            if mod in self.external:
                raise ValueError(f"extra-mod '{mod}' also appears in external mods")

        self.display = [item.lower() for item in self.display]
        self.extensions = list(set(self.extensions) | set(self.fpp_extensions))
        self.exclude_dir.append(self.output_dir)
        self.extra_mods.update(INTRINSIC_MODS)

        # Check that none of the docmarks are the same
        docmarks = ["docmark", "predocmark", "docmark_alt", "predocmark_alt"]
        for first, second in combinations(docmarks, 2):
            first_mark = getattr(self, first)
            second_mark = getattr(self, second)
            if first_mark == second_mark != "":
                raise ValueError(
                    f"{first} ('{first_mark}') and {second} ('{second_mark}') are the same"
                )

        if isinstance(self.extra_filetypes, list):
            try:
                self.extra_filetypes = {
                    filetype.extension: filetype for filetype in self.extra_filetypes
                }
            except AttributeError:
                file_types = [
                    ExtraFileType(**file_type) for file_type in self.extra_filetypes
                ]
                self.extra_filetypes = {
                    file_type.extension: file_type for file_type in file_types
                }

    @classmethod
    def from_markdown_metadata(cls, meta: Dict[str, Any], parent: Optional[str] = None):
        return cls(**convert_types_from_metapreprocessor(cls, meta, parent))

    def normalise_paths(self, directory=None):
        if directory is None:
            directory = Path.cwd()
        self.directory = Path(directory).absolute()
        field_types = get_type_hints(ProjectSettings)

        if self.favicon == FAVICON_PATH:
            self.favicon = Path(__file__).parent / FAVICON_PATH

        if self.md_base_dir == Path("."):
            self.md_base_dir = self.directory

        for key, value in asdict(self).items():
            if value is None:
                continue

            default_type = field_types[key]

            if is_same_type(default_type, List[Path]):
                value = getattr(self, key)
                setattr(self, key, [normalise_path(self.directory, v) for v in value])

            if is_same_type(default_type, Path):
                setattr(self, key, normalise_path(self.directory, value))

        if self.relative:
            self.project_url = self.output_dir


def load_toml_settings(directory: PathLike) -> Optional[ProjectSettings]:
    """Load Ford settings from ``fpm.toml`` file in ``directory``

    Settings should be in ``[extra.ford]`` table
    """

    filename = Path(directory) / "fpm.toml"

    if not filename.is_file():
        return None

    with open(filename, "rb") as f:
        settings = tomllib.load(f)

    if "extra" not in settings:
        return None

    if "ford" not in settings["extra"]:
        return None

    print(f"Reading Ford options from {filename.absolute()}")
    try:
        return ProjectSettings(**settings["extra"]["ford"])
    except ValueError as e:
        raise ValueError(f"Error parsing settings from '{filename}': {e}")


def load_markdown_settings(
    directory: PathLike, project_file: str, filename: Optional[str] = None
) -> Tuple[ProjectSettings, str]:
    settings, project_lines = meta_preprocessor(project_file)
    settings = convert_types_from_metapreprocessor(ProjectSettings, settings, filename)

    # Workaround for file inclusion in metadata
    for option, value in settings.items():
        if isinstance(value, str) and MD_INCLUDE_RE.match(value):
            warn(
                "Including other files in project file metadata is deprecated and "
                "will stop working in a future release.\n"
                f"    {option}: {value}",
            )
            md_base_dir = settings.get("md_base_dir", directory)
            configs = MarkdownInclude({"base_path": str(md_base_dir)}).getConfigs()
            include_preprocessor = IncludePreprocessor(None, configs)
            settings[option] = "\n".join(include_preprocessor.run(value.splitlines()))

    try:
        return ProjectSettings.from_markdown_metadata(settings), "\n".join(
            project_lines
        )
    except ValueError as e:
        raise ValueError(f"Error parsing settings from '{filename}': {e}")


def convert_setting(default_type: Type, key: str, value: Any) -> Any:
    """Convert an individual value's type to be consistent with a given dataclass for the
    given key."""

    if is_same_type(default_type, type(value)):
        # already correct type
        return value
    if is_same_type(default_type, list):
        return [value]
    elif is_same_type(default_type, bool):
        return convert_to_bool(key, value)
    elif is_same_type(default_type, int):
        return int(value[0])
    elif (
        is_same_type(default_type, str) or is_same_type(default_type, Path)
    ) and isinstance(value, list):
        return "\n".join(value)
    elif (get_origin(default_type) is dict) and not isinstance(value, dict):
        resvalue = value
        if isinstance(value, str):
            resvalue = [value]

        # Get rid of any empty strings
        resvalue = [v for v in resvalue if v]

        if get_args(default_type) == (str, ExtraFileType):
            file_types = [ExtraFileType.from_string(string) for string in resvalue]
            return {file_type.extension: file_type for file_type in file_types}
        else:
            sep = OPTION_SEPARATORS[key]
            return _parse_to_dict(resvalue, name=key, sep=sep)

    # Nothing special to do
    return value


def convert_types_from_metapreprocessor(
    cls: Type, settings: Dict[str, Any], parent: Optional[str] = None
):
    """Convert a dict's value's types to be consistent with a given dataclass"""

    field_types = get_type_hints(cls)

    keys_to_drop = []

    for key, value in settings.items():
        try:
            default_type = field_types[key]
        except KeyError:
            prefix = f"In '{parent}': " if parent is not None else ""
            warn(f"{prefix}Ignoring unknown Ford metadata key {key!r}")
            keys_to_drop.append(key)
            continue

        settings[key] = convert_setting(default_type, key, value)

    for key in keys_to_drop:
        settings.pop(key)

    return settings


def convert_types_from_commandarguments(
    settings: ProjectSettings, cargs: Dict[str, Any], parent: Optional[str] = None
):
    """Convert the cargs dict's value's types to be consistent with a given dataclass
    if set and override them in settings.
    """

    field_types = get_type_hints(type(settings))

    for key, value in cargs.items():
        if value is not None:
            if key in field_types:
                setattr(settings, key, convert_setting(field_types[key], key, value))
            else:
                setattr(settings, key, value)

    return settings


def _parse_to_dict(string_list: List[str], name: str, sep: str) -> Dict[str, str]:
    """Parse a list of strings of form "key = value" into a dict

    Parameters
    ----------
    string_list : List[str]
        List of strings to parse
    name : str
        Name in parent settings object, only used for error message
    sep: str
        Separator between key and value

    """

    result = {}
    for string in string_list:
        try:
            key, value = string.split(sep, 1)
        except ValueError:
            raise RuntimeError(
                f"Error setting option {name!r}: expected '{sep}' in {string!r}"
            )

        # Remove extraneous quotes for the URL options
        if sep == ":":
            value.strip().strip(r"\"'")

        result[key.strip()] = value.strip()
    return result


@dataclass
class EntitySettings:
    author: Optional[str] = None
    category: Optional[str] = None
    copy_subdir: List[Path] = field(default_factory=list)
    date: Optional[str] = None
    deprecated: bool = False
    display: List[str] = field(default_factory=list)
    graph: bool = ProjectSettings.graph
    graph_maxdepth: int = ProjectSettings.graph_maxdepth
    graph_maxnodes: int = ProjectSettings.graph_maxnodes
    license: Optional[str] = None
    num_lines: Optional[int] = None
    ordered_subpage: List[str] = field(default_factory=list)
    proc_internals: bool = ProjectSettings.proc_internals
    since: Optional[str] = None
    source: bool = ProjectSettings.source
    summary: Optional[str] = None
    title: Optional[str] = None
    version: Optional[str] = None

    @classmethod
    def from_markdown_metadata(cls, meta: Dict[str, Any], parent: Optional[str] = None):
        return cls(**convert_types_from_metapreprocessor(cls, meta, parent))

    @classmethod
    def from_project_settings(cls, project_settings: ProjectSettings):
        """Inherit entity-specific settings from project-level settings"""
        return cls(
            graph=project_settings.graph,
            graph_maxdepth=project_settings.graph_maxdepth,
            graph_maxnodes=project_settings.graph_maxnodes,
            proc_internals=project_settings.proc_internals,
            source=project_settings.source,
        )

    def update(self, metadata: Dict[str, Any], parent: Optional[str] = None) -> None:
        """Update self with values from a dict"""
        current_settings = asdict(self)
        current_settings.update(
            convert_types_from_metapreprocessor(type(self), metadata, parent)
        )
        for key, value in current_settings.items():
            setattr(self, key, value)



from ford.sourceform import (
    FortranSourceFile,
    FortranModule,
    FortranBase,
    parse_type,
    ParsedType,
    line_to_variables,
    GenericSource,
)
from ford.fortran_project import find_used_modules
from ford import ProjectSettings
from ford._markdown import MetaMarkdown

from dataclasses import dataclass, field
from typing import Union, List, Optional
from itertools import chain
from textwrap import dedent

import pytest


class FakeProject:
    def __init__(self, procedures=None):
        self.procedures = procedures or []


@pytest.fixture
def parse_fortran_file(copy_fortran_file):
    def parse_file(data, **kwargs):
        filename = copy_fortran_file(data)
        settings = ProjectSettings(**kwargs)
        return FortranSourceFile(str(filename), settings)

    return parse_file


def test_extends(parse_fortran_file):
    """Check that types can be extended"""

    data = """\
    program foo
    !! base type
    type :: base
    end type base

    !! derived type
    type, extends(base) :: derived
    end type

    !! derived type but capitalised
    type, EXTENDS(base) :: derived_capital
    end type

    end program foo
    """

    fortran_type = parse_fortran_file(data)

    assert len(fortran_type.programs) == 1

    program = fortran_type.programs[0]

    assert len(program.types) == 3
    assert program.types[1].extends == "base"
    assert program.types[2].extends == "base"


def test_type_visibility_attributes(parse_fortran_file):
    """Check that we can set visibility attributes on types, #388"""

    data = """\
    module default_public
      type no_attrs
      end type no_attrs

      type, public :: public_attr
      end type public_attr

      type, private :: private_attr
      end type private_attr

      type, public :: public_attr_private_components
        private
      end type public_attr

      type, private :: private_attr_public_components
        public
      end type private_attr
    end module default_public

    module default_private
      private

      type no_attrs
      end type no_attrs

      type, public :: public_attr
      end type public_attr

      type, private :: private_attr
      end type private_attr

      type, public :: public_attr_private_components
        private
      end type public_attr

      type, private :: private_attr_public_components
        public
      end type private_attr
    end module default_private
    """

    source = parse_fortran_file(data)
    public_no_attrs = source.modules[0].types[0]
    public_public_attr = source.modules[0].types[1]
    public_private_attr = source.modules[0].types[2]
    public_public_attr_components = source.modules[0].types[3]
    public_private_attr_components = source.modules[0].types[4]

    assert public_no_attrs.permission == "public"
    assert public_public_attr.permission == "public"
    assert public_private_attr.permission == "private"
    assert public_public_attr_components.permission == "public"
    assert public_private_attr_components.permission == "private"

    private_no_attrs = source.modules[1].types[0]
    private_public_attr = source.modules[1].types[1]
    private_private_attr = source.modules[1].types[2]
    private_public_attr_components = source.modules[1].types[3]
    private_private_attr_components = source.modules[1].types[4]

    assert private_no_attrs.permission == "private"
    assert private_public_attr.permission == "public"
    assert private_private_attr.permission == "private"
    assert private_public_attr_components.permission == "public"
    assert private_private_attr_components.permission == "private"


def test_submodule_procedure_contains(parse_fortran_file):
    """Check that submodule procedures can have 'contains' statements"""

    data = """\
    module foo_m
      implicit none
      interface
        module subroutine foo()
          implicit none
        end subroutine
      end interface
    end module

    submodule(foo_m) foo_s
      implicit none
    contains
      module procedure foo
      contains
        subroutine bar()
        end subroutine
      end procedure
    end submodule
    """

    fortran_type = parse_fortran_file(data)

    assert len(fortran_type.modules) == 1
    assert len(fortran_type.submodules) == 1
    submodule = fortran_type.submodules[0]
    assert len(submodule.modprocedures) == 1
    module_procedure = submodule.modprocedures[0]
    assert len(module_procedure.subroutines) == 1


def test_backslash_in_character_string(parse_fortran_file):
    """Bad escape crash #296"""

    data = r"""\
    module test_module
    character(len=*),parameter,public:: q = '(?)'
    character(len=*),parameter,public:: a  = '\a'
    character(len=*),parameter,public:: b  = '\b'
    character(len=*),parameter,public:: c  = '\c'
    end module test_module
    """

    source = parse_fortran_file(data)
    module = source.modules[0]

    expected_variables = {"q": r"'(?)'", "a": r"'\a'", "b": r"'\b'", "c": r"'\c'"}

    for variable in module.variables:
        assert variable.initial == expected_variables[variable.name]


def test_sync_images_in_submodule_procedure(parse_fortran_file):
    """Crash on sync images inside module procedure in submodule #237"""

    data = """\
    module stuff
      interface
        module subroutine foo()
        end subroutine
      end interface
    end module

    submodule(stuff) sub_stuff
      implicit none
    contains
      module procedure foo
        sync images(1)
      end procedure
    end submodule
    """

    parse_fortran_file(data)


def test_function_and_subroutine_call_on_same_line(parse_fortran_file):
    """Regex does not check for nested calls #256"""

    data = """\
    program test
    call bar(foo())
    contains
    integer function foo()
    end function foo
    subroutine bar(thing)
      integer, intent(in) :: thing
    end subroutine bar
    end program test
    """

    fortran_file = parse_fortran_file(data)
    program = fortran_file.programs[0]
    assert len(program.calls) == 2
    expected_calls = {"bar", "foo"}
    assert set(call[-1] for call in program.calls) == expected_calls


@pytest.mark.parametrize(
    ["call_segment", "expected"],
    [
        (
            """
            USE m_baz, ONLY: t_bar
            USE m_foo, ONLY: t_baz

            TYPE(t_bar) :: v_bar
            TYPE(t_baz) :: var
            var = v_bar%p_foo()
            """,
            ["p_foo"],
        ),
        (
            """
            USE m_baz, ONLY: t_bar

            TYPE(t_bar) :: v_bar
            INTEGER :: var
            var = v_bar%v_foo(1)
            """,
            [],
        ),
        (
            """
            USE m_baz, ONLY: t_bar

            TYPE(t_bar) :: v_bar
            INTEGER :: var
            var = [v_bar%v_foo(1)]
            """,
            [],
        ),
        (
            """
            USE m_baz, ONLY: t_bar

            TYPE(t_bar) :: v_bar
            INTEGER, DIMENSION(:), ALLOCATABLE :: var
            var = v_bar%v_baz%p_baz()
            """,
            ["p_baz"],
        ),
        (
            """
            USE m_baz, ONLY: t_bar
            USE m_foo, ONLY: t_baz

            TYPE(t_bar) :: v_bar
            TYPE(t_baz) :: var
            var = v_bar%t_foo%p_foo()
            """,
            ["p_foo"],
        ),
        (
            """
            USE m_baz, ONLY: t_bar
            USE m_foo, ONLY: t_baz

            TYPE(t_bar) :: v_bar
            TYPE(t_baz) :: var(1)
            var = [v_bar%t_foo%p_foo()]
            """,
            ["p_foo"],
        ),
        (
            """
            USE m_baz, ONLY: t_bar

            TYPE(t_bar), DIMENSION(2) :: v_bar
            INTEGER :: var
            var = v_bar(1)%v_baz%v_faz(1)
            """,
            [],
        ),
        (
            """
            USE m_baz, ONLY: t_bar

            TYPE(t_bar), DIMENSION(2) :: v_bar
            call v_bar(1)%renamed()
            """,
            ["renamed"],
        ),
        (
            """
            USE m_baz, ONLY: t_bar

            TYPE(t_bar), DIMENSION(2) :: v_bar
            call v_bar ( 1 ) % p_bar
            """,
            ["p_bar"],
        ),
        (
            """
            USE m_baz, ONLY: t_bar

            TYPE(t_bar) :: v_bar
            call v_bar % p_bar ( v_bar % v_baz % p_baz () )
            """,
            ["p_bar", "p_baz"],
        ),
        (
            """
            USE m_foo, ONLY: t_baz
            USE m_baz, ONLY: p_buz

            TYPE(t_baz) :: var_baz
            call p_buz(var_baz%p_baz())
            """,
            ["p_baz", "p_buz"],
        ),
        (
            """
            USE m_foo, ONLY: t_baz

            TYPE(t_baz) :: var_baz
            write(*,*) var_baz%p_baz()
            """,
            ["p_baz"],
        ),
        (
            """
            USE m_baz, ONLY: t_bar
            TYPE(t_bar) :: v_bar
            ASSOCIATE (tmp => v_bar)
                CALL tmp%p_bar()
            END ASSOCIATE
            """,
            ["p_bar"],
        ),
        (
            """
            USE m_baz, ONLY: t_bar
            TYPE(t_bar) :: v_bar
            INTEGER, DIMENSION(2) :: var
            ASSOCIATE (tmp => v_bar%v_baz)
                var = tmp%p_baz()
            END ASSOCIATE
            """,
            ["p_baz"],
        ),
        (
            """
            USE m_baz, ONLY: t_bar
            TYPE(t_bar) :: v_bar
            INTEGER :: var
            ASSOCIATE (tmp => v_bar%v_baz)
                ASSOCIATE (tmp2 => tmp%p_baz())
                END ASSOCIATE
            END ASSOCIATE
            """,
            ["p_baz"],
        ),
        (
            """
            USE m_baz, ONLY: t_bar
            TYPE(t_bar) :: v_bar
            INTEGER, DIMENSION(1) :: var_arr
            ASSOCIATE (tmp => v_bar%v_baz)
                ASSOCIATE (tmp => tmp%v_faz)
                    var_arr = tmp
                END ASSOCIATE
                var_arr = tmp%p_baz()
            END ASSOCIATE
            """,
            ["p_baz"],
        ),
        (
            """
            USE m_baz, ONLY: t_bar
            INTEGER :: var = 10
            ASSOCIATE (tmp => var)
                var = tmp
            END ASSOCIATE
            """,
            [],
        ),
        (
            """
            USE m_baz, ONLY: t_bar
            TYPE(t_bar) :: v_bar
            ASSOCIATE (tmp => v_bar%p_foo())
                PRINT *, tmp
            END ASSOCIATE
            """,
            ["p_foo"],
        ),
        (
            """
            USE m_baz, ONLY: t_bar
            TYPE(t_bar) :: v_bar
            INTEGER, DIMENSION(2) :: var_arr
            ASSOCIATE (tmp => v_bar, arr => var_arr)
                CALL tmp%p_bar()
                arr = tmp%v_baz%p_baz()
            END ASSOCIATE
            """,
            ["p_bar", "p_baz"],
        ),
        (
            """
            USE m_baz, ONLY: t_bar
            TYPE(t_bar) :: v_bar
            INTEGER, DIMENSION(2) :: var_arr
            ASSOCIATE (tmp => v_bar)
                ASSOCIATE (tmp2 => tmp%v_baz%p_baz())
                END ASSOCIATE
            END ASSOCIATE
            """,
            ["p_baz"],
        ),
        (
            """
            ASSOCIATE (tmp => p_faz(p_fuz(1)))
            END ASSOCIATE
            """,
            ["p_faz", "p_fuz"],
        ),
        (
            """
            USE m_baz, ONLY: t_unknown_parent
            TYPE(t_unknown_parent)
            call t_unknown_parent%known_method()
            """,
            ["known_method"],
        ),
    ],
)
def test_type_chain_function_and_subroutine_calls(
    parse_fortran_file, call_segment, expected
):
    data = f"""\
    MODULE m_foo
        IMPLICIT NONE
        TYPE :: t_baz
            INTEGER, DIMENSION(:), ALLOCATABLE :: v_faz
        CONTAINS
            PROCEDURE :: p_baz
        END TYPE t_baz

        CONTAINS

        FUNCTION p_baz(self) RESULT(ret_val)
            CLASS(t_baz), INTENT(IN) :: self
            INTEGER, DIMENSION(:), ALLOCATABLE :: ret_val
        END FUNCTION p_baz

    END MODULE m_foo

    MODULE m_bar
        TYPE :: t_foo
            INTEGER, DIMENSION(:), ALLOCATABLE :: v_foo
        CONTAINS
            PROCEDURE :: p_foo
        END TYPE t_foo

        CONTAINS

        FUNCTION p_foo(self) RESULT(ret_val)
            USE m_foo, ONLY: t_baz

            CLASS(t_foo), INTENT(IN) :: self
            TYPE(t_baz) :: ret_val
        END FUNCTION p_foo

    END MODULE m_bar

    MODULE m_baz

        USE m_foo, ONLY: t_baz
        USE m_bar, ONLY: t_foo
        USE unknown_module
        TYPE, EXTENDS(t_foo) :: t_bar
            TYPE(t_baz) :: v_baz
        CONTAINS
            PROCEDURE :: p_bar
            PROCEDURE :: renamed => p_bar
        END TYPE t_bar

        TYPE, EXTENDS(unknown_type) :: t_unknown_parent
        CONTAINS
            PROCEDURE, NOPASS :: known_method
        END TYPE

        CONTAINS

        SUBROUTINE p_bar(self)
            CLASS(t_bar), INTENT(IN) :: self
        END SUBROUTINE p_bar

        SUBROUTINE p_buz(var_int)
            INTEGER, DIMENSION(2), INTENT(IN) :: var_int
        END SUBROUTINE p_buz

        SUBROUTINE known_method()
        END SUBROUTINE

    END MODULE m_baz

    MODULE m_main

        CONTAINS
        
        FUNCTION p_faz(arg) RESULT(ret_val)
            INTEGER, INTENT(IN) :: arg
            INTEGER :: ret_val
        END FUNCTION p_faz

        FUNCTION p_fuz(arg) RESULT(ret_val)
            INTEGER, INTENT(IN) :: arg
            INTEGER :: ret_val
        END FUNCTION p_fuz
        
        SUBROUTINE main
            {call_segment}
        END SUBROUTINE main

    END MODULE m_main
    """

    fortran_file = parse_fortran_file(data)
    fp = FakeProject()
    modules = {module.name: module for module in fortran_file.modules}
    for module in modules.values():
        find_used_modules(module, modules.values(), [], [])

    # correlation order is important
    modules["m_foo"].correlate(fp)
    modules["m_bar"].correlate(fp)
    modules["m_baz"].correlate(fp)
    modules["m_main"].correlate(fp)

    main_subroutines = {sub.name: sub for sub in modules["m_main"].subroutines}
    calls = main_subroutines["main"].calls

    assert len(calls) == len(expected)

    calls_sorted = sorted(calls, key=lambda x: getattr(x, "name", x))
    expected_sorted = sorted(expected)
    for call, expected_name in zip(calls_sorted, expected_sorted):
        assert isinstance(call, FortranBase)

        assert call.name == expected_name


def test_call_in_module_procedure(parse_fortran_file):
    data = """\
    module foo
      type :: nuz
      contains
        procedure :: cor
      end type nuz
      interface
        module subroutine bar()
        end subroutine bar
      end interface
    contains
      function cor(this_) result (ret_val)
        class (nuz), intent(in) :: this_
        real :: ret_val
      end function cor
    end module foo

    submodule(foo) baz
    contains
      module procedure bar
        type (nuz) :: var
        real :: val
        val = var%cor()
      end procedure bar
    end submodule baz
    """
    expected = ["cor"]

    fortran_file = parse_fortran_file(data, display=["public", "protected", "private"])
    fp = FakeProject()
    modules = {
        module.name: module
        for module in chain(fortran_file.modules, fortran_file.submodules)
    }
    for module in modules.values():
        find_used_modules(module, modules.values(), [], [])

    # correlation order is important
    modules["foo"].correlate(fp)
    modules["baz"].correlate(fp)

    main_subroutines = {sub.name: sub for sub in modules["baz"].modprocedures}
    calls = main_subroutines["bar"].calls

    assert len(calls) == len(expected)

    calls_sorted = sorted(calls, key=lambda x: getattr(x, "name", x))
    expected_sorted = sorted(expected)
    for call, expected_name in zip(calls_sorted, expected_sorted):
        assert isinstance(call, FortranBase)

        assert call.name == expected_name


def test_submodule_private_var_call(parse_fortran_file):
    data = """\
    module foo
      type :: nuz
        integer :: var
      end type nuz

      type(nuz), dimension(2), private :: pri_var

      interface
        module subroutine bar()
        end subroutine bar
      end interface
    end module foo

    submodule(foo) baz
    contains
      module procedure bar
        integer :: val
        val = pri_var(1)%var
      end procedure bar
    end submodule baz
    """

    fortran_file = parse_fortran_file(data)
    fp = FakeProject()
    modules = {
        module.name: module
        for module in chain(fortran_file.modules, fortran_file.submodules)
    }
    for module in modules.values():
        find_used_modules(module, modules.values(), [], [])

    # correlation order is important
    modules["foo"].correlate(fp)
    modules["baz"].correlate(fp)

    main_subroutines = {sub.name: sub for sub in modules["baz"].modprocedures}
    calls = main_subroutines["bar"].calls

    assert calls == []


def test_internal_proc_arg_var_call(parse_fortran_file):
    data = """\
    module foo
      type :: nuz
        integer :: var
      end type nuz
      type(nuz), dimension(2), private :: pri_var
    contains
      subroutine bar(nuz_var)
        class(nuz), dimension(2) :: nuz_var
      contains
        subroutine baz
          integer :: val
          val = nuz_var(1)%var
        end subroutine baz
      end subroutine bar
    end module foo
    """

    fortran_file = parse_fortran_file(data)
    fp = FakeProject()
    modules = {module.name: module for module in fortran_file.modules}
    for module in modules.values():
        find_used_modules(module, modules.values(), [], [])

    # correlation order is important
    modules["foo"].correlate(fp)

    calls = modules["foo"].subroutines[0].subroutines[0].calls

    assert calls == []


def test_component_access(parse_fortran_file):
    data = """\
    module mod1
        integer :: anotherVar(20)
        type typeOne
            integer :: ivar(10)
        end type typeOne
        type(typeOne) :: One
    end module mod1

    module mod2
        type typeTwo
            integer :: ivar(10)
        end type typeTwo
        type(typeTwo) :: Two
    end module mod2

    subroutine with_space
        use mod2
        integer :: a
        a = 3
        Two% ivar(:) = a
    end subroutine with_space

    program main
        integer :: i, j
        integer :: zzz(5)

        type typeThree
            integer :: ivar(10)
        end type typeThree
        type(typeThree) :: Three

        call with_space()
        call without_space()
        anotherVar(3) = i
        j = zzz(3)

        Three% ivar(3) = 7
    end program
    """

    fortran_file = parse_fortran_file(data)

    expected_variables = {"i", "j", "zzz", "Three"}
    actual_variables = {var.name for var in fortran_file.programs[0].variables}
    assert actual_variables == expected_variables


def test_format_statement(parse_fortran_file):
    """No function calls in `format` statements are allowed, so don't
    confuse them with format specifiers. Issue #350"""

    data = """\
    program test_format_statement
      implicit none
      write (*, 300)
    300 format (/1X, 44('-'), ' Begin of test2 Calculation ', 33('-')//)
    end program test_format_statement
    """

    fortran_file = parse_fortran_file(data)
    assert fortran_file.programs[0].calls == []


def test_enumerator_with_kind(parse_fortran_file):
    """Checking enumerators with specified kind, issue #293"""

    data = """\
    module some_enums
      use, intrinsic :: iso_fortran_env, only : int32
      enum, bind(c)
        enumerator :: item1, item2
        enumerator :: list1 = 100_int32, list2
        enumerator :: fixed_item1 = 0, fixed_item2
      end enum
    end module some_enums
    """

    fortran_file = parse_fortran_file(data)
    enum = fortran_file.modules[0].enums[0]
    assert enum.variables[0].name == "item1"
    assert enum.variables[0].initial == 0
    assert enum.variables[1].name == "item2"
    assert enum.variables[1].initial == 1
    assert enum.variables[2].name == "list1"
    assert enum.variables[2].initial == "100_int32"
    assert enum.variables[3].name == "list2"
    assert enum.variables[3].initial == 101
    assert enum.variables[4].name == "fixed_item1"
    assert enum.variables[4].initial == "0"
    assert enum.variables[5].name == "fixed_item2"
    assert enum.variables[5].initial == 1


class FakeModule(FortranModule):
    def __init__(
        self, procedures: dict, interfaces: dict, types: dict, variables: dict
    ):
        self.pub_procs = procedures
        self.pub_absints = interfaces
        self.pub_types = types
        self.pub_vars = variables


def test_module_get_used_entities_all():
    mod_procedures = {"subroutine": "some subroutine"}
    mod_interfaces = {"abstract": "interface"}
    mod_types = {"mytype": "some type"}
    mod_variables = {"x": "some var"}

    module = FakeModule(mod_procedures, mod_interfaces, mod_types, mod_variables)

    procedures, interfaces, types, variables = module.get_used_entities("")

    assert procedures == mod_procedures
    assert interfaces == mod_interfaces
    assert types == mod_types
    assert variables == mod_variables


def test_module_get_used_entities_some():
    mod_procedures = {"subroutine": "some subroutine"}
    mod_interfaces = {"abstract": "interface"}
    mod_types = {"mytype": "some type"}
    mod_variables = {"x": "some var", "y": "some other var"}

    module = FakeModule(mod_procedures, mod_interfaces, mod_types, mod_variables)

    procedures, interfaces, types, variables = module.get_used_entities(
        ", only: x, subroutine"
    )

    assert procedures == mod_procedures
    assert interfaces == {}
    assert types == {}
    assert variables == {"x": mod_variables["x"]}


def test_module_get_used_entities_rename():
    mod_procedures = {"subroutine": "some subroutine"}
    mod_interfaces = {"abstract": "interface"}
    mod_types = {"mytype": "some type"}
    mod_variables = {"x": "some var", "y": "some other var"}

    module = FakeModule(mod_procedures, mod_interfaces, mod_types, mod_variables)

    procedures, interfaces, types, variables = module.get_used_entities(
        ", only: x, y => subroutine"
    )

    assert procedures == {"y": mod_procedures["subroutine"]}
    assert interfaces == {}
    assert types == {}
    assert variables == {"x": mod_variables["x"]}


def test_module_default_access(parse_fortran_file):
    data = """\
    module default_access
      ! No access keyword
      integer :: int_public, int_private
      private :: int_private
      real :: real_public
      real, private :: real_private

      type :: type_public
        complex :: component_public
        complex, private :: component_private
      end type type_public

      type :: type_private
        character(len=1) :: string_public
        character(len=1), private :: string_private
      end type type_private

      private :: sub_private, func_private, type_private

    contains
      subroutine sub_public
      end subroutine sub_public

      subroutine sub_private
      end subroutine sub_private

      integer function func_public()
      end function func_public

      integer function func_private()
      end function func_private
    end module default_access
    """

    fortran_file = parse_fortran_file(data)
    fortran_file.modules[0].correlate(FakeProject())

    assert set(fortran_file.modules[0].all_procs.keys()) == {
        "sub_public",
        "func_public",
        "sub_private",
        "func_private",
    }
    assert set(fortran_file.modules[0].pub_procs.keys()) == {
        "sub_public",
        "func_public",
    }
    assert set(fortran_file.modules[0].all_types.keys()) == {
        "type_public",
        "type_private",
    }
    assert set(fortran_file.modules[0].pub_types.keys()) == {
        "type_public",
    }
    assert set(fortran_file.modules[0].all_vars.keys()) == {
        "int_public",
        "int_private",
        "real_public",
        "real_private",
    }
    assert set(fortran_file.modules[0].pub_vars.keys()) == {
        "int_public",
        "real_public",
    }


def test_module_public_access(parse_fortran_file):
    data = """\
    module public_access
      public
      integer :: int_public, int_private
      private :: int_private
      real :: real_public
      real, private :: real_private

      type :: type_public
        complex :: component_public
        complex, private :: component_private
      end type type_public

      type :: type_private
        character(len=1) :: string_public
        character(len=1), private :: string_private
      end type type_private

      private :: sub_private, func_private, type_private

    contains
      subroutine sub_public
      end subroutine sub_public

      subroutine sub_private
      end subroutine sub_private

      integer function func_public()
      end function func_public

      integer function func_private()
      end function func_private
    end module public_access
    """

    fortran_file = parse_fortran_file(data)
    fortran_file.modules[0].correlate(FakeProject())

    assert set(fortran_file.modules[0].all_procs.keys()) == {
        "sub_public",
        "func_public",
        "sub_private",
        "func_private",
    }
    assert set(fortran_file.modules[0].pub_procs.keys()) == {
        "sub_public",
        "func_public",
    }
    assert set(fortran_file.modules[0].all_types.keys()) == {
        "type_public",
        "type_private",
    }
    assert set(fortran_file.modules[0].pub_types.keys()) == {
        "type_public",
    }
    assert set(fortran_file.modules[0].all_vars.keys()) == {
        "int_public",
        "int_private",
        "real_public",
        "real_private",
    }
    assert set(fortran_file.modules[0].pub_vars.keys()) == {
        "int_public",
        "real_public",
    }


def test_module_private_access(parse_fortran_file):
    data = """\
    module private_access
      private
      integer :: int_public, int_private
      public :: int_public
      real :: real_private
      real, public :: real_public

      type :: type_public
        complex :: component_public
        complex, private :: component_private
      end type type_public

      type :: type_private
        character(len=1) :: string_public
        character(len=1), private :: string_private
      end type type_private

      public :: sub_public, func_public, type_public

    contains
      subroutine sub_public
      end subroutine sub_public

      subroutine sub_private
      end subroutine sub_private

      integer function func_public()
      end function func_public

      integer function func_private()
      end function func_private
    end module private_access
    """

    fortran_file = parse_fortran_file(data)
    fortran_file.modules[0].correlate(FakeProject())

    assert set(fortran_file.modules[0].all_procs.keys()) == {
        "sub_public",
        "func_public",
        "sub_private",
        "func_private",
    }
    assert set(fortran_file.modules[0].pub_procs.keys()) == {
        "sub_public",
        "func_public",
    }
    assert set(fortran_file.modules[0].all_types.keys()) == {
        "type_public",
        "type_private",
    }
    assert set(fortran_file.modules[0].pub_types.keys()) == {
        "type_public",
    }
    assert set(fortran_file.modules[0].all_vars.keys()) == {
        "int_public",
        "int_private",
        "real_public",
        "real_private",
    }
    assert set(fortran_file.modules[0].pub_vars.keys()) == {
        "int_public",
        "real_public",
    }


def test_module_procedure_case(parse_fortran_file):
    """Check that submodule procedures in interface blocks are parsed correctly. Issue #353"""
    data = """\
    module a
      implicit none
      interface
        MODULE SUBROUTINE square( x )
          integer, intent(inout):: x
        END SUBROUTINE square
        module subroutine cube( x )
          integer, intent(inout):: x
        end subroutine cube
        MODULE FUNCTION square_func( x )
          integer, intent(in):: x
        END FUNCTION square_func
        module function cube_func( x )
          integer, intent(inout):: x
        end function cube_func
      end interface
    end module a

    submodule (a) b
      implicit none
    contains
      MODULE PROCEDURE square
        x = x * x
      END PROCEDURE square
      module PROCEDURE cube
        x = x * x * x
      END PROCEDURE cube
      MODULE PROCEDURE square_func
        square_func = x * x
      END PROCEDURE square_func
      module procedure cube_func
        cube_func = x * x * x
      end procedure cube_func
    end submodule b
    """

    fortran_file = parse_fortran_file(data)
    module = fortran_file.modules[0]
    assert len(module.interfaces) == 4
    assert module.interfaces[0].procedure.module
    assert module.interfaces[1].procedure.module
    assert module.interfaces[2].procedure.module
    assert module.interfaces[3].procedure.module


def test_submodule_ancestors(parse_fortran_file):
    """Check that submodule ancestors and parents are correctly identified"""

    data = """\
    module mod_a
    end module mod_a

    submodule (mod_a) mod_b
    end submodule mod_b

    submodule (mod_a) mod_c
    end submodule mod_c

    submodule (mod_a:mod_c) mod_d
    end submodule mod_d
    """

    fortran_file = parse_fortran_file(data)

    mod_b = fortran_file.submodules[0]
    mod_c = fortran_file.submodules[1]
    mod_d = fortran_file.submodules[2]

    assert mod_b.parent_submodule is None
    assert mod_b.ancestor_module == "mod_a"

    assert mod_c.parent_submodule is None
    assert mod_c.ancestor_module == "mod_a"

    assert mod_d.parent_submodule == "mod_c"
    assert mod_d.ancestor_module == "mod_a"


@pytest.mark.parametrize(
    ["variable_decl", "expected"],
    [
        ("integer i", ParsedType("integer", "i")),
        ("integer :: i", ParsedType("integer", ":: i")),
        ("integer ( int32 ) :: i", ParsedType("integer", ":: i", "int32")),
        ("real r", ParsedType("real", "r")),
        ("real(real64) r", ParsedType("real", "r", "real64")),
        ("REAL( KIND  =  8) :: r, x, y", ParsedType("real", ":: r, x, y", "8")),
        ("REAL( 8 ) :: r, x, y", ParsedType("real", ":: r, x, y", "8")),
        ("complex*16 znum", ParsedType("complex", "znum", "16")),
        (
            "character(len=*) :: string",
            ParsedType("character", ":: string", strlen="*"),
        ),
        (
            "character(len=:) :: string",
            ParsedType("character", ":: string", strlen=":"),
        ),
        ("character(12) :: string", ParsedType("character", ":: string", strlen="12")),
        (
            "character(var) :: string",
            ParsedType("character", ":: string", strlen="var"),
        ),
        ("character :: string", ParsedType("character", ":: string", strlen="1")),
        (
            "character(LEN=12) :: string",
            ParsedType("character", ":: string", strlen="12"),
        ),
        (
            "CHARACTER(KIND= kind('0') ,  len =12) :: string",
            ParsedType("character", ":: string", kind='kind("a")', strlen="12"),
        ),
        (
            "CHARACTER(KIND=kanji,  len =12) :: string",
            ParsedType("character", ":: string", kind="kanji", strlen="12"),
        ),
        (
            "CHARACTER(  len =   12,KIND=kanji) :: string",
            ParsedType("character", ":: string", kind="kanji", strlen="12"),
        ),
        (
            "CHARACTER( 12,kanji ) :: string",
            ParsedType("character", ":: string", kind="kanji", strlen="12"),
        ),
        (
            "CHARACTER(  kind=    kanji) :: string",
            ParsedType("character", ":: string", kind="kanji", strlen="1"),
        ),
        ("double PRECISION dp", ParsedType("double precision", "dp")),
        ("DOUBLE   complex dc", ParsedType("double complex", "dc")),
        (
            "type(something) :: thing",
            ParsedType("type", ":: thing", proto=["something", ""]),
        ),
        (
            "type(character(kind=kanji, len=10)) :: thing",
            ParsedType("type", ":: thing", proto=["character", "kind=kanji,len=10"]),
        ),
        (
            "class(foo) :: thing",
            ParsedType("class", ":: thing", proto=["foo", ""]),
        ),
        (
            "procedure(bar) :: thing",
            ParsedType("procedure", ":: thing", proto=["bar", ""]),
        ),
        ("Vec :: vector", ParsedType("vec", ":: vector")),
        ("Mat :: matrix", ParsedType("mat", ":: matrix")),
    ],
)
def test_parse_type(variable_decl, expected):
    # Tokeniser will have previously replaced strings with index into
    # this list
    capture_strings = ['"a"']
    result = parse_type(variable_decl, capture_strings, ["Vec", "Mat"])
    assert result.vartype == expected.vartype
    assert result.kind == expected.kind
    assert result.strlen == expected.strlen
    assert result.proto == expected.proto
    assert result.rest == expected.rest


class FakeSource:
    def __init__(self):
        self.text = iter(["end subroutine", "end module"])

    def __next__(self):
        return next(self.text)

    def pass_back(self, line):
        pass


@dataclass
class FakeParent:
    strings: List[str] = field(default_factory=lambda: ['"Hello"', "'World'"])
    settings: ProjectSettings = field(default_factory=ProjectSettings)
    obj: str = "module"
    parent = None
    display: List[str] = field(default_factory=list)
    _to_be_markdowned: List[FortranBase] = field(default_factory=list)
    doxy_dict = {}


def _make_list_str() -> List[str]:
    """This is just to stop mypy complaining for ``attribs`` below"""
    return []


@dataclass
class FakeVariable:
    name: str
    vartype: str
    parent: Optional[FakeParent] = field(default_factory=FakeParent)
    attribs: Optional[List[str]] = field(default_factory=_make_list_str)
    intent: str = ""
    optional: bool = False
    permission: str = "public"
    parameter: bool = False
    kind: Optional[str] = None
    strlen: Optional[str] = None
    proto: Union[None, str, List[str]] = None
    doc_list: List[str] = field(default_factory=list)
    points: bool = False
    initial: Optional[str] = None


@pytest.mark.parametrize(
    ["line", "expected_variables"],
    [
        ("integer foo", [FakeVariable("foo", "integer")]),
        ("integer :: foo", [FakeVariable("foo", "integer")]),
        (
            "real :: foo, bar",
            [FakeVariable("foo", "real"), FakeVariable("bar", "real")],
        ),
        (
            "real, allocatable :: foo",
            [FakeVariable("foo", "real", attribs=["allocatable"])],
        ),
        (
            "integer, intent ( in  out)::zing",
            [FakeVariable("zing", "integer", intent="inout")],
        ),
        (
            "real(real64), optional, intent(in) :: foo, bar",
            [
                FakeVariable("foo", "real", kind="real64", optional=True, intent="in"),
                FakeVariable("bar", "real", kind="real64", optional=True, intent="in"),
            ],
        ),
        (
            "character ( len = 24 , kind = 4), parameter :: char = '0', far = '1'",
            [
                FakeVariable(
                    "char",
                    "character",
                    strlen="24",
                    kind="4",
                    parameter=True,
                    initial='"Hello"',
                ),
                FakeVariable(
                    "far",
                    "character",
                    strlen="24",
                    kind="4",
                    parameter=True,
                    initial="'World'",
                ),
            ],
        ),
        (
            "procedure(foo) :: bar",
            [FakeVariable("bar", "procedure", proto=["foo", ""])],
        ),
        (
            "type(foo) :: bar = 42",
            [FakeVariable("bar", "type", proto=["foo", ""], initial="42")],
        ),
        (
            "class(foo) :: var1, var2",
            [
                FakeVariable("var1", "class", proto=["foo", ""]),
                FakeVariable("var2", "class", proto=["foo", ""]),
            ],
        ),
        (
            "class(*) :: polymorphic",
            [FakeVariable("polymorphic", "class", proto=["*", ""])],
        ),
    ],
)
def test_line_to_variable(line, expected_variables):
    variables = line_to_variables(FakeSource(), line, "public", FakeParent())
    attributes = expected_variables[0].__dict__
    attributes.pop("parent")

    for variable, expected in zip(variables, expected_variables):
        for attr in attributes:
            variable_attr = getattr(variable, attr)
            expected_attr = getattr(expected, attr)
            assert variable_attr == expected_attr, attr

    if len(expected_variables) > 1:
        for attr in attributes:
            attribute = getattr(variable, attr)
            if not isinstance(attribute, list):
                continue
            proto_ids = [id(getattr(variable, attr)) for variable in variables]
            assert len(proto_ids) == len(set(proto_ids)), attr


def test_markdown_header_bug286(parse_fortran_file):
    """Check that markdown headers work, issue #286"""
    data = """\
    module myModule
    contains
      subroutine printSquare(x)
        !! ## My Header
        !! This should be one section, but the header doesn't work
        integer, intent(in) :: x
        write(*,*) x*x
      end subroutine printSquare
    end module myModule
    """

    fortran_file = parse_fortran_file(data)
    md = MetaMarkdown()

    subroutine = fortran_file.modules[0].subroutines[0]
    subroutine.markdown(md)

    assert subroutine.doc.startswith("<h2>My Header</h2>")


def test_markdown_codeblocks_bug286(parse_fortran_file):
    """Check that markdown codeblocks work, issue #287"""
    data = """\
    module myModule
    contains
      subroutine printSquare(x)
        !! This codeblock should not be inline:
        !! ```
        !! printSquare(4)
        !! ```
        integer, intent(in) :: x
        write(*,*) x*x
      end subroutine printSquare
    end module myModule
    """

    fortran_file = parse_fortran_file(data)
    md = MetaMarkdown()

    subroutine = fortran_file.modules[0].subroutines[0]
    subroutine.markdown(md)

    assert "<code>printSquare(4)" in subroutine.doc
    assert "<div" in subroutine.doc


def test_markdown_meta_reset(parse_fortran_file):
    """Check that markdown metadata is reset between entities"""
    data = """\
    module myModule
      !! version: 0.1.0
    contains
      subroutine printSquare(x)
        !! author: Test name
        integer, intent(in) :: x
        write(*,*) x*x
      end subroutine printSquare
      !> @author Lucie Forrest
      subroutine printCube(x)
        integer, intent(in) :: x
        write(*,*) x*x*x
      end subroutine printCube
    end module myModule
    """

    fortran_file = parse_fortran_file(data)
    md = MetaMarkdown()
    module = fortran_file.modules[0]
    module.markdown(md)
    assert module.meta.version == "0.1.0"
    assert module.subroutines[0].meta.author == "Test name"
    assert module.subroutines[1].meta.author == "Lucie Forrest"


def test_multiline_attributes(parse_fortran_file):
    """Check that specifying attributes over multiple lines works"""

    data = """\
    program prog
      real x
      dimension x(:)
      allocatable x
      integer, allocatable, dimension(:) :: y
      complex z
      allocatable z(:)

      allocate(x(1), y(1), z(1))
    end program prog
    """

    fortran_file = parse_fortran_file(data)
    prog = fortran_file.programs[0]

    for variable in prog.variables:
        assert (
            "allocatable" in variable.attribs
        ), f"Missing 'allocatable' in '{variable}' attributes"
        assert (
            variable.dimension == "(:)" or "dimension(:)" in variable.attribs
        ), f"Wrong dimension for '{variable}'"


def test_markdown_source_meta(parse_fortran_file):
    """Check that specifying 'source' in the procedure meta block is processed"""

    data = """\
    subroutine with_source
    !! source: true
    !!
    !! some docs
    end subroutine with_source
    """

    md = MetaMarkdown()

    fortran_file = parse_fortran_file(data)
    subroutine = fortran_file.subroutines[0]
    subroutine.markdown(md)

    assert subroutine.meta.source is True
    assert "with_source" in subroutine.src


def test_markdown_source_settings(parse_fortran_file):
    """Check that specifying 'source' in the settings works"""

    data = """\
    subroutine with_source
    !! some docs
    end subroutine with_source
    """

    md = MetaMarkdown()
    fortran_file = parse_fortran_file(data, source=True)
    subroutine = fortran_file.subroutines[0]
    subroutine.markdown(md)

    assert subroutine.meta.source is True
    assert "with_source" in subroutine.src


@pytest.mark.parametrize(
    ["snippet", "expected_error", "expected_name"],
    (
        (
            "program foo\n contains\n contains",
            "Multiple CONTAINS",
            "foo",
        ),
        (
            "program foo\n interface bar\n contains",
            "Unexpected CONTAINS",
            "bar",
        ),
        ("end", "END statement", "test.f90"),
        ("program foo\n module procedure bar", "Unexpected MODULE PROCEDURE", "foo"),
        ("program foo\n module bar", "Unexpected MODULE", "foo"),
        (
            "program foo\n submodule (foo) bar \n end program foo",
            "Unexpected SUBMODULE",
            "program 'foo'",
        ),
        ("program foo\n program bar", "Unexpected PROGRAM", "foo"),
        (
            "program foo\n end program foo\n program bar \n end program bar",
            "Multiple PROGRAM",
            "test.f90",
        ),
        (
            "program foo\n subroutine bar \n end subroutine \n end program",
            "Unexpected SUBROUTINE",
            "program 'foo'",
        ),
        (
            "program foo\n integer function bar() \n end function bar\n end program foo",
            "Unexpected FUNCTION",
            "program 'foo'",
        ),
    ),
)
def test_bad_parses(snippet, expected_error, expected_name, parse_fortran_file):
    with pytest.raises(ValueError) as e:
        parse_fortran_file(snippet, dbg=False)

    assert expected_error in e.value.args[0]
    assert expected_name in e.value.args[0]


def test_routine_iterator(parse_fortran_file):
    data = """\
    module foo
      interface
        module subroutine modsub1()
        end subroutine modsub1
        module subroutine modsub2()
        end subroutine modsub2
        module integer function modfunc1()
        end function modfunc1
        module integer function modfunc2()
        end function modfunc2
      end interface
    contains
      subroutine sub1()
      end subroutine sub1
      subroutine sub2()
      end subroutine sub2
      integer function func1()
      end function func1
      integer function func2()
      end function func2
    end module foo

    submodule (foo) bar
    contains
      module subroutine modsub1
      end subroutine modsub1
      module subroutine modsub2
      end subroutine modsub2
      module procedure modfunc1
      end procedure modfunc1
      module procedure modfunc2
      end procedure modfunc2

      subroutine sub3()
      end subroutine sub3
      subroutine sub4()
      end subroutine sub4
      integer function func3()
      end function func3
      integer function func4()
      end function func4
    end submodule bar
    """

    fortran_file = parse_fortran_file(data)

    module = fortran_file.modules[0]
    assert sorted([proc.name for proc in module.routines]) == [
        "func1",
        "func2",
        "sub1",
        "sub2",
    ]

    submodule = fortran_file.submodules[0]
    assert sorted([proc.name for proc in submodule.routines]) == [
        "func3",
        "func4",
        "modfunc1",
        "modfunc2",
        "modsub1",
        "modsub2",
        "sub3",
        "sub4",
    ]


def test_type_component_permissions(parse_fortran_file):
    data = """\
    module default_access
      private
      type :: type_default
        complex :: component_public
        complex, private :: component_private
      contains
        procedure :: sub_public
        procedure, private :: sub_private
      end type type_default

      type, public :: type_public
        public
        complex :: component_public
        complex, private :: component_private
      contains
        public
        procedure :: sub_public
        procedure, private :: sub_private
      end type type_public

      type :: type_private
        private
        character(len=1), public :: string_public
        character(len=1) :: string_private
      contains
        private
        procedure, public :: sub_public
        procedure :: sub_private
      end type type_private

      type :: type_public_private
        public
        character(len=1) :: string_public
        character(len=1), private :: string_private
      contains
        private
        procedure, public :: sub_public
        procedure :: sub_private
      end type type_public_private

      type :: type_private_public
        private
        character(len=1), public :: string_public
        character(len=1) :: string_private
      contains
        public
        procedure :: sub_public
        procedure, private :: sub_private
      end type type_private_public

      public :: type_default, type_private_public, type_public_private
    contains
      subroutine sub_public
      end subroutine sub_public

      subroutine sub_private
      end subroutine sub_private
    end module default_access
    """

    fortran_file = parse_fortran_file(data)
    fortran_file.modules[0].correlate(FakeProject())

    for ftype in fortran_file.modules[0].types:
        assert (
            ftype.variables[0].permission == "public"
        ), f"{ftype.name}::{ftype.variables[0].name}"
        assert (
            ftype.variables[1].permission == "private"
        ), f"{ftype.name}::{ftype.variables[1].name}"
        assert (
            ftype.boundprocs[0].permission == "public"
        ), f"{ftype.name}::{ftype.boundprocs[0].name}"
        assert (
            ftype.boundprocs[1].permission == "private"
        ), f"{ftype.name}::{ftype.boundprocs[1].name}"


def test_variable_formatting(parse_fortran_file):
    data = """\
    module foo_m
      character(kind=kind('a'), len=4), dimension(:, :), allocatable :: multidimension_string
      type :: bar
      end type bar
      type(bar), parameter :: something = bar()
    contains
      type(bar) function quux()
      end function quux
    end module foo_m
    """

    fortran_file = parse_fortran_file(data)
    fortran_file.modules[0].correlate(FakeProject())
    variable0 = fortran_file.modules[0].variables[0]
    variable1 = fortran_file.modules[0].variables[1]

    assert variable0.full_type == "character(kind=kind('a'), len=4)"
    assert (
        variable0.full_declaration
        == "character(kind=kind('a'), len=4), dimension(:, :), allocatable"
    )
    assert variable1.full_type == "type(bar)"
    assert variable1.full_declaration == "type(bar), parameter"

    function = fortran_file.modules[0].functions[0]
    assert function.retvar.full_declaration == "type(bar)"


def test_url(parse_fortran_file):
    data = """\
    program prog_foo
      integer :: int_foo
    contains
      subroutine sub_foo
      end subroutine sub_foo
      function func_foo()
      end function func_foo
    end program prog_foo

    module mod_foo
      real :: real_foo
      interface inter_foo
        module procedure foo1, foo2
      end interface inter_foo
      type :: foo_t
        integer :: int_bar
      end type
      enum, bind(C)
        enumerator :: red = 4, blue = 9
        enumerator :: yellow
      end enum
    contains
      subroutine foo1()
      end subroutine foo1
      subroutine foo2(x)
        integer :: x
      end subroutine foo2
    end module mod_foo

    submodule (mod_foo) submod_foo
    end submodule submod_foo
    """

    fortran_file = parse_fortran_file(data)
    program = fortran_file.programs[0]
    module = fortran_file.modules[0]
    submodule = fortran_file.submodules[0]

    assert program.get_dir() == "program"
    assert module.get_dir() == "module"
    assert submodule.get_dir() == "module"
    assert program.subroutines[0].get_dir() == "proc"
    assert program.functions[0].get_dir() == "proc"
    assert module.subroutines[0].get_dir() == "proc"
    assert module.interfaces[0].get_dir() == "interface"
    assert program.variables[0].get_dir() is None
    assert module.variables[0].get_dir() is None
    assert module.enums[0].get_dir() is None

    assert program.full_url.endswith("program/prog_foo.html")
    assert module.full_url.endswith("module/mod_foo.html")
    assert submodule.full_url.endswith("module/submod_foo.html")
    assert program.subroutines[0].full_url.endswith("proc/sub_foo.html")
    assert program.functions[0].full_url.endswith("proc/func_foo.html")
    assert module.subroutines[0].full_url.endswith("proc/foo1.html")
    assert module.interfaces[0].full_url.endswith("interface/inter_foo.html")
    assert program.variables[0].full_url.endswith(
        "program/prog_foo.html#variable-int_foo"
    )
    assert module.variables[0].full_url.endswith(
        "module/mod_foo.html#variable-real_foo"
    )


def test_single_character_interface(parse_fortran_file):
    data = """\
    module a
      interface b !! some comment
        module procedure c
      end interface b
    end module a
    """
    fortran_file = parse_fortran_file(data)
    assert fortran_file.modules[0].interfaces[0].name == "b"
    assert fortran_file.modules[0].interfaces[0].doc_list == [" some comment"]


def test_module_procedure_in_module(parse_fortran_file):
    data = """\
    module foo_mod
      interface
        module subroutine quaxx
        end subroutine quaxx
      end interface
    contains
      module procedure quaxx
        print*, "implementation"
      end procedure
    end module foo_mod
    """

    fortran_file = parse_fortran_file(data)
    module = fortran_file.modules[0]
    module.correlate(FakeProject())

    interface = module.interfaces[0]
    assert interface.name == "quaxx"
    modproc = module.modprocedures[0]

    assert interface.procedure.module == modproc
    assert modproc.module == interface


def test_module_interface_same_name_as_interface(parse_fortran_file):
    data = """\
    module foo_m
      interface foo
        module function foo() result(bar)
          logical bar
        end function
      end interface
    contains
      module procedure foo
        bar = .true.
      end procedure
    end module
    """

    fortran_file = parse_fortran_file(data)
    module = fortran_file.modules[0]
    module.correlate(FakeProject())

    interface = module.interfaces[0]
    assert interface.name == "foo"

    modproc = module.modprocedures[0]
    assert modproc.name == "foo"


def test_procedure_pointer(parse_fortran_file):
    data = """\
    module foo
      abstract interface
        integer pure function unary_f_t(n)
          implicit none
          integer, intent(in) :: n
        end function
      end interface

      private

      procedure(unary_f_t), pointer, public :: unary_f => null()

      interface generic_unary_f
        procedure unary_f
      end interface
    end module
    """

    fortran_file = parse_fortran_file(data)
    module = fortran_file.modules[0]
    module.correlate(FakeProject())
    assert len(module.interfaces[0].modprocs) == 0
    assert module.interfaces[0].variables[0].name == "unary_f"


def test_block_data(parse_fortran_file):
    data = """\
    block data name
      !! Block data docstring
      common /name/ foo
      !! Common block docstring

      character*31 foo(1024)
      !! Variable docstring

      data foo /'a', 'b', 'c', 'd', 'e', 1019*'0'/
    end
    """

    fortran_file = parse_fortran_file(data)
    blockdata = fortran_file.blockdata[0]

    assert blockdata.name == "name"
    assert blockdata.doc_list[0].strip() == "Block data docstring"
    assert len(blockdata.common) == 1
    assert blockdata.common[0].doc_list[0].strip() == "Common block docstring"
    assert len(blockdata.variables) == 1
    assert blockdata.variables[0].doc_list[0].strip() == "Variable docstring"


def test_subroutine_empty_args(parse_fortran_file):
    data = """\
    subroutine foo (    )
    end subroutine foo
    """

    fortran_file = parse_fortran_file(data)
    subroutine = fortran_file.subroutines[0]
    assert subroutine.args == []


def test_subroutine_whitespace(parse_fortran_file):
    data = """\
    subroutine foo (  a,b,    c,d  )
      integer :: a, b, c, d
    end subroutine foo
    """

    fortran_file = parse_fortran_file(data)
    subroutine = fortran_file.subroutines[0]
    arg_names = [arg.name for arg in subroutine.args]
    assert arg_names == ["a", "b", "c", "d"]


def test_function_empty_args(parse_fortran_file):
    data = """\
    integer function foo (    )
    end function foo
    """

    fortran_file = parse_fortran_file(data)
    function = fortran_file.functions[0]
    assert function.args == []


def test_function_whitespace(parse_fortran_file):
    data = """\
    integer function foo (  a,b,    c,d  )
      integer :: a, b, c, d
    end function foo
    """

    fortran_file = parse_fortran_file(data)
    function = fortran_file.functions[0]
    arg_names = [arg.name for arg in function.args]
    assert arg_names == ["a", "b", "c", "d"]


def test_bind_name_subroutine(parse_fortran_file):
    data = """\
    subroutine init() bind(C, name="c_init")
    end subroutine init
    """

    fortran_file = parse_fortran_file(data)
    subroutine = fortran_file.subroutines[0]

    assert subroutine.bindC == 'C, name="c_init"'


def test_bind_name_function(parse_fortran_file):
    data = """\
    integer function foo() bind(C, name="c_foo")
    end function foo
    """

    fortran_file = parse_fortran_file(data)
    function = fortran_file.functions[0]

    assert function.bindC == 'C, name="c_foo"'


def test_generic_bound_procedure(parse_fortran_file):
    data = """\
    module subdomain_m
      type subdomain_t
      contains
        procedure no_colon
        procedure :: colon
        generic :: operator(+) => no_colon, colon
      end type
      interface
        module function no_colon(lhs, rhs)
          class(subdomain_t), intent(in) :: lhs
          integer, intent(in) :: rhs
          type(subdomain_t) total
        end function
        module function colon(lhs, rhs)
          class(subdomain_t), intent(in) :: lhs, rhs
          type(subdomain_t) total
        end function
      end interface
    end module
    """

    fortran_file = parse_fortran_file(data)
    fortran_type = fortran_file.modules[0].types[0]

    expected_names = sorted(["no_colon", "colon", "operator(+)"])
    bound_proc_names = sorted([proc.name for proc in fortran_type.boundprocs])
    assert bound_proc_names == expected_names


def test_submodule_procedure_calls(parse_fortran_file):
    """Check that calls inside submodule procedures are correctly correlated"""

    data = """\
    module foo_m
      implicit none
      interface
        module function foo1(start, end) result(res)
          integer, intent(in) :: start, end
          integer :: res
        end function
      end interface
    end module

    submodule(foo_m) foo_s
      implicit none
    contains
      integer function bar(start, end)
        integer, intent(in) :: start, end
        bar = end - start
      end function

      module procedure foo1
        res = bar(start, end)
      end procedure
    end submodule
    """

    fortran_file = parse_fortran_file(data)
    fortran_file.modules[0].correlate(FakeProject())
    submodule = fortran_file.submodules[0]
    submodule.correlate(FakeProject())

    assert submodule.modprocedures[0].calls[0] == submodule.functions[0]


def test_namelist(parse_fortran_file):
    data = """\
    module mod_a
      integer :: var_a
    end module mod_a
    module mod_b
      use mod_a
      integer :: var_b
    end module mod_b

    program prog
      integer :: var_c
    contains
      subroutine sub(var_d)
        use mod_b
        integer, intent(in) :: var_d
        integer :: var_e
        namelist /namelist_a/ var_a, var_b, var_c, var_d, var_e
        !! namelist docstring
      end subroutine sub
    end program prog
    """
    fortran_file = parse_fortran_file(data)
    namelist = fortran_file.programs[0].subroutines[0].namelists[0]
    assert namelist.name == "namelist_a"

    expected_names = sorted(["var_a", "var_b", "var_c", "var_d", "var_e"])
    output_names = sorted(namelist.variables)
    assert output_names == expected_names

    assert namelist.doc_list == [" namelist docstring"]


def test_namelist_correlate(parse_fortran_file):
    data = """\
    program prog
      integer :: var_c
    contains
      subroutine sub(var_b)
        integer, intent(in) :: var_b
        integer :: var_a
        namelist /namelist_a/ var_a, var_b, var_c
      end subroutine sub
    end program prog
    """
    fortran_file = parse_fortran_file(data)
    fortran_file.programs[0].correlate(FakeProject())
    namelist = fortran_file.programs[0].subroutines[0].namelists[0]
    expected_names = sorted(["var_a", "var_b", "var_c"])
    output_names = sorted([variables.name for variables in namelist.variables])
    assert output_names == expected_names


def test_generic_source(tmp_path):
    data = """\
    #! docmark
    #* docmark_alt
    #> predocmark
    #| predocmark_alt
    """

    filename = tmp_path / "generic_source.sh"
    filename.write_text(dedent(data))

    settings = ProjectSettings(
        extra_filetypes=[{"extension": "sh", "comment": "#"}],
    )
    source = GenericSource(filename, settings)
    expected_docs = [
        "docmark",
        "docmark_alt",
        "predocmark",
        "predocmark_alt",
    ]

    assert source.doc_list == expected_docs


def test_type_bound_procedure_formatting(parse_fortran_file):
    data = """\
    module ford_example_type_mod
      type, abstract, public :: example_type
      contains
        procedure :: say_hello => example_type_say
        procedure, private, nopass :: say_int
        procedure, private, nopass :: say_real
        generic :: say_number => say_int, say_real
        procedure(say_interface), deferred :: say
        final :: example_type_finalise
      end type example_type

      interface
        subroutine say_interface(self)
          import say_type_base
          class(say_type_base), intent(inout) :: self
        end subroutine say_interface
      end interface

    contains

      subroutine example_type_say(self)
        class(example_type), intent(inout) :: self
      end subroutine example_type_say

      subroutine example_type_finalise(self)
        type(example_type), intent(inout) :: self
      end subroutine example_type_finalise

      subroutine say_int(int)
        integer, intent(in) :: int
      end subroutine say_int

      subroutine say_real(r)
        real, intent(in) :: r
      end subroutine say_real
    end module ford_example_type_mod
    """

    example_type = parse_fortran_file(data).modules[0].types[0]

    assert example_type.boundprocs[0].full_declaration == "procedure, public"
    assert example_type.boundprocs[1].full_declaration == "procedure, private, nopass"
    assert example_type.boundprocs[2].full_declaration == "procedure, private, nopass"
    assert example_type.boundprocs[3].full_declaration == "generic, public"
    assert (
        example_type.boundprocs[4].full_declaration
        == "procedure(say_interface), public, deferred"
    )


def test_type_num_lines(parse_fortran_file):
    data = """\
    module ford_example_type_mod
      type, abstract, public :: example_type
      contains
        procedure :: say_hello => example_type_say
        procedure, private, nopass :: say_int
        procedure, private, nopass :: say_real
        generic :: say_number => say_int, say_real, say_hello
        procedure(say_interface), deferred :: say
      end type example_type
      interface
        subroutine say_interface(self)
          import say_type_base
          class(say_type_base), intent(inout) :: self
        end subroutine say_interface
      end interface
    contains
      subroutine example_type_say(self)
        class(example_type), intent(inout) :: self
      end subroutine example_type_say
      subroutine say_int(int)
        integer, intent(in) :: int
      end subroutine say_int
      subroutine say_real(r)
        real, intent(in) :: r
      end subroutine say_real
    end module ford_example_type_mod
    """

    project = FakeProject()
    module = parse_fortran_file(data).modules[0]
    module.correlate(project)

    example_type = module.types[0]
    assert example_type.num_lines == 8
    assert example_type.num_lines_all == 8 + 9


def test_associate_array(parse_fortran_file):
    data = """\
    subroutine test()
      associate(phi => [1,2], theta => (/ 3, 4 /))
      end associate
    end subroutine test"""

    # Just check we can parse ok
    parse_fortran_file(data)


def test_blocks_with_type(parse_fortran_file):
    data = """\
    module foo
    contains
      subroutine sub1()
        block
          type :: t1
          end type t1
        end block
      end subroutine sub1
    end module foo
    """

    source = parse_fortran_file(data)
    module = source.modules[0]
    assert len(module.subroutines) == 1


def test_no_space_after_character_type(parse_fortran_file):
    data = """\
    CHARACTER(LEN=250)FUNCTION FirstWord( sString ) RESULT( sRes )
        CHARACTER(LEN=250)sString
    END FUNCTION FirstWord
    """

    source = parse_fortran_file(data)
    function = source.functions[0]
    assert function.name.lower() == "firstword"


def test_summary_handling_bug703(parse_fortran_file):
    """Check that `summary` works, PR #703"""
    data = """\
    !> summary: Lorem ipsum dolor sit amet, consectetur adipiscing elit.
    !>          Fusce ultrices tortor et felis tempus vehicula.
    !>          Nulla gravida, magna ut pharetra.
    !>
    !> Full description
    module test
    end module myModule
    """

    fortran_file = parse_fortran_file(data)
    md = MetaMarkdown()

    module = fortran_file.modules[0]
    module.markdown(md)

    assert "Lorem ipsum" in module.meta.summary


def test_doxygen_parameters(parse_fortran_file):
    data = """\
    !> @param stuff some comment
    !> Normal Comment
    !> @param[in] stuff_2 Doxygen comment
    !> @param     stuff_3 Comment should not show
    module a
      integer, intent(in) :: stuff
      integer, intent(in) :: stuff_2
      !! FORD comment
    end module a
    module b
      integer, intent(in) :: stuff_3
      !! Should only capture this comment for stuff_3
    end module a
    """
    fortran_file = parse_fortran_file(data)
    module_a = fortran_file.modules[0]
    module_b = fortran_file.modules[1]
    assert module_a.doc_list == [" Normal Comment"]
    # single doxygen comment
    assert module_a.variables[0].doc_list == [" some comment"]
    # doxygen comment and a FORD comment
    assert module_a.variables[1].doc_list == [" FORD comment", " Doxygen comment"]
    # doxygen comments are only shown for comments directly above the subroutine
    assert module_b.variables[0].doc_list == [
        " Should only capture this comment for stuff_3"
    ]


def test_no_doxygen_parameters(parse_fortran_file):
    data = """\
    !> Normal Comment
    !> @param stuff_2 Doxygen comment
    module a
      integer, intent(in) :: stuff
      integer, intent(in) :: stuff_2
      !! FORD comment
    end module a
    module b
      integer, intent(in) :: stuff_3
      !!comment for stuff_3
    end module a
    """
    fortran_file = parse_fortran_file(data, doxygen=False)
    module_a = fortran_file.modules[0]
    module_b = fortran_file.modules[1]
    assert module_a.doc_list == [" Normal Comment", " @param stuff_2 Doxygen comment"]
    assert module_a.variables[0].doc_list == []
    assert module_a.variables[1].doc_list == [" FORD comment"]
    # doxygen comments are only shown for comments directly above the subroutine
    assert module_b.variables[0].doc_list == ["comment for stuff_3"]


def test_doxygen_metadata_translation(parse_fortran_file):
    data = """\
    !> @brief summary bit
    !> Main details
    module a
    end module a
    """
    fortran_file = parse_fortran_file(data)
    module = fortran_file.modules[0]
    assert module.meta.summary == "summary bit"
    assert module.doc_list == [" Main details"]


def test_doxygen_multiple_metadata(parse_fortran_file):
    data = """\
    !> @brief summary bit
    !> @author Foo Shamaloo
    !> Main details
    !> @version 3.14
    module a
    end module a
    """
    fortran_file = parse_fortran_file(data)
    module = fortran_file.modules[0]
    assert module.meta.summary == "summary bit"
    assert module.meta.author == "Foo Shamaloo"
    assert module.meta.version == "3.14"
    assert module.doc_list == [" Main details"]


def test_doxygen_ok_on_source_file(parse_fortran_file):
    data = """\
    !! @brief Source file
    module a
    end module a
    """
    fortran_file = parse_fortran_file(data)
    assert fortran_file.meta.summary == "Source file"


def test_doxygen_see_link(parse_fortran_file):
    data = """\
    !> @see b
    module a
    end module a
    !> @see a with a description
    module b
    end module b
    """
    fortran_file = parse_fortran_file(data)
    assert fortran_file.modules[1].doc_list[0].strip() == "[[a]] with a description"
    assert fortran_file.modules[0].doc_list[0].strip() == "[[b]]"


def test_doxygen_eating_admonitions_bug717(parse_fortran_file):
    """Ensure the doxygen parser doesn't convert our admonitions"""

    data = """\
    !> @note this should remain
    module test
    end module test
    """

    fortran_file = parse_fortran_file(data)
    md = MetaMarkdown()

    module = fortran_file.modules[0]
    module.markdown(md)

    # This is tied to the format of the admonitions class
    assert 'div class="alert alert-info"' in module.doc



# -*- coding: utf-8 -*-
#
#  pagetree.py
#  This file is part of FORD.
#
#  Copyright 2015 Christopher MacMackin <cmacmackin@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from __future__ import annotations

import os
from pathlib import Path
from typing import List, Optional, Sequence
from textwrap import dedent
from ford.console import warn
from ford.utils import meta_preprocessor, ProgressBar
from ford.settings import EntitySettings, ProjectSettings
from ford._markdown import MetaMarkdown


class PageNode:
    """
    Object representing a page in a tree of pages and subpages.
    """

    def __init__(
        self,
        md: MetaMarkdown,
        path: Path,
        output_dir: Path,
        proj_copy_subdir: Sequence[Path],
        parent: Optional[PageNode],
        encoding: str = "utf-8",
    ):
        meta, text = meta_preprocessor(dedent(path.read_text(encoding)))
        self.meta = EntitySettings.from_markdown_metadata(meta, path.stem)
        self.base_url = md.base_url

        if self.meta.title is None:
            raise ValueError(f"Page '{path}' has no title metadata")

        self.title = self.meta.title
        self.author = self.meta.author
        self.date = self.meta.date

        # index.md is the main page, it should not be added by the user in the subpage lists.
        self.ordered_subpages = [
            x for x in self.meta.ordered_subpage if x != "index.md"
        ]

        # set list of directory names that are to be copied along without
        # containing an index.md itself.
        #   first priority is the copy_dir option in the *.md file
        #   if this option is not set in the file fall back to the global
        #   project settings
        self.copy_subdir = self.meta.copy_subdir or proj_copy_subdir
        self.parent = parent
        self.subpages: List[PageNode] = []
        self.files: List[Path] = []
        self.filename = Path(path.stem)
        if self.parent:
            self.hierarchy: List[PageNode] = self.parent.hierarchy + [self.parent]
            self.topdir: Path = self.parent.topdir
            self.location = Path(os.path.relpath(path.parent, self.topdir))
            self.topnode: PageNode = self.parent.topnode
        else:
            self.hierarchy = []
            self.topdir = path.parent
            self.location = Path()
            self.topnode = self

        output_path = output_dir / "page" / self.path.parent
        self.contents = md.reset().convert("\n".join(text), path=output_path.resolve())

    @property
    def path(self):
        return self.location / self.filename.with_suffix(".html")

    @property
    def url(self):
        return self.base_url / "page" / self.path

    def __str__(self):
        return f"<a href='{self.url}'>{self.title}</a>"

    def __iter__(self):
        retlist = [self]
        for sp in self.subpages:
            retlist.extend(list(sp.__iter__()))
        return iter(retlist)


def is_excluded_dir(path: Path, settings: ProjectSettings) -> bool:
    """Should `path` be excluded from the pagetree?"""
    paths_to_exclude: set[Path] = {
        dir
        for dir in [settings.graph_dir, settings.media_dir, settings.output_dir]
        if dir is not None
    }
    paths_to_exclude.update(settings.html_template_dir)
    paths_to_exclude.update(settings.src_dir)

    return path in paths_to_exclude


def get_page_tree(
    topdir: Path,
    proj_copy_subdir: Sequence[Path],
    output_dir: Path,
    md: MetaMarkdown,
    settings: ProjectSettings,
    progress: Optional[ProgressBar] = None,
    parent=None,
    encoding: str = "utf-8",
):
    # In python 3.6 or newer, the normal dict is guaranteed to be ordered.
    # However, to keep compatibility with older versions I use OrderedDict.
    # I will use this later to remove duplicates from a list in a short way.
    from collections import OrderedDict

    # look for files within topdir
    index_file = topdir / "index.md"

    if not index_file.exists():
        # Don't warn if this is a Ford input directory (media_dir, graph_dir, etc)
        if not is_excluded_dir(topdir, settings):
            warn(
                f"Skipping creating page for '{topdir}' becase it does not contain 'index.md'"
            )
        return None

    if progress is not None:
        progress.set_current(os.path.relpath(index_file))
    # process index.md
    try:
        node = PageNode(md, index_file, output_dir, proj_copy_subdir, parent, encoding)
    except Exception as e:
        warn(f"Error parsing {index_file.relative_to(topdir)}.\n\t{e.args[0]}")
        return None

    filelist = sorted(os.listdir(topdir))
    filelist.remove("index.md")

    if node.ordered_subpages:
        # Merge user given files and all files in folder, removing duplicates.
        mergedfilelist = list(OrderedDict.fromkeys(node.ordered_subpages + filelist))
    else:
        mergedfilelist = filelist

    for name in mergedfilelist:
        if name[0] == ".":
            continue
        if name[-1] == "~":
            continue

        filename = topdir / name
        if progress is not None:
            progress.set_current(os.path.relpath(filename))

        if not filename.exists():
            raise ValueError(f"Requested page file '{filename}' does not exist")

        if filename.is_dir():
            # recurse into subdirectories
            if parent and name in parent.copy_subdir:
                continue

            if subnode := get_page_tree(
                filename,
                proj_copy_subdir,
                output_dir,
                md,
                settings,
                progress,
                node,
                encoding,
            ):
                node.subpages.append(subnode)
        elif filename.suffix == ".md":
            # process subpages
            try:
                node.subpages.append(
                    PageNode(md, filename, output_dir, proj_copy_subdir, node, encoding)
                )
            except ValueError as e:
                warn(f"Error parsing '{filename}'.\n\t{e.args[0]}")
                continue
        else:
            node.files.append(Path(name))

    return node



import shutil
import sys
import os
import pathlib
from urllib.parse import urlparse
import json
from typing import Dict, Any

import ford

from bs4 import BeautifulSoup
import pytest

from conftest import chdir

REMOTE_TYPE_JSON: Dict[str, Any] = {
    "name": "remote_type",
    "external_url": "./type/remote_type.html",
    "obj": "type",
    "extends": None,
    "variables": [
        {
            "name": "cptr",
            "external_url": "./type/config.html#variable-cptr",
            "obj": "variable",
            "vartype": "type",
            "permission": "public",
        },
    ],
    "boundprocs": [],
    "permission": "public",
}

REMOTE_MODULES_JSON = [
    {
        "name": "remote_module",
        "external_url": "./module/remote_module.html",
        "obj": "module",
        "pub_procs": {
            "remote_sub": {
                "name": "remote_sub",
                "external_url": "./proc/remote_sub.html",
                "obj": "proc",
                "proctype": "Subroutine",
                "functions": [],
                "subroutines": [],
                "interfaces": [],
                "absinterfaces": [],
                "types": [],
                "variables": [],
            },
        },
        "pub_absints": {},
        "pub_types": {"remote_type": REMOTE_TYPE_JSON},
        "pub_vars": {},
        "functions": [],
        "subroutines": [
            {
                "name": "remote_sub",
                "external_url": "./proc/remote_sub.html",
                "obj": "proc",
                "proctype": "Subroutine",
                "functions": [],
                "subroutines": [],
                "interfaces": [],
                "absinterfaces": [],
                "types": [],
                "variables": [],
            }
        ],
        "interfaces": [],
        "absinterfaces": [],
        "types": [REMOTE_TYPE_JSON],
        "variables": [],
    }
]


class MockResponse:
    @staticmethod
    def read():
        return json.dumps(REMOTE_MODULES_JSON).encode("utf-8")


@pytest.fixture(scope="module")
def monkeymodule(request):
    """pytest won't let us use function-scope fixtures in module-scope
    fixtures, so we need to reimplement this with module scope"""
    mpatch = pytest.MonkeyPatch()
    yield mpatch
    mpatch.undo()


@pytest.fixture(scope="module")
def external_project(tmp_path_factory, monkeymodule):
    """Generate the documentation for an "external" project and then
    for a "top level" one that uses the first.

    A remote external project is simulated through a mocked `urlopen`
    which returns `REMOTE_MODULES_JSON`

    """

    this_dir = pathlib.Path(__file__).parent
    path = tmp_path_factory.getbasetemp() / "external_project"
    shutil.copytree(this_dir / "../../test_data/external_project", path)

    external_project = path / "external_project"
    top_level_project = path / "top_level_project"

    # Run FORD in the two projects
    # First project has "externalize: True" and will generate JSON dump
    with monkeymodule.context() as m, chdir(external_project):
        m.setattr(sys, "argv", ["ford", "doc.md"])
        ford.run()

    def mock_open(*args, **kwargs):
        return MockResponse()

    # Second project uses JSON from first to link to external modules
    with monkeymodule.context() as m, chdir(top_level_project):
        m.setattr(sys, "argv", ["ford", "doc.md"])
        m.setattr(ford.external_project, "urlopen", mock_open)
        ford.run()

    # Make sure we're in a directory where relative paths won't
    # resolve correctly
    os.chdir("/")

    return top_level_project, external_project


def test_external_project(external_project):
    """Check that we can build external projects and get the links correct"""

    top_level_project, _ = external_project

    # Read generated HTML
    program_dir = top_level_project / "doc/program"
    with open(program_dir / "top_level.html", "r") as f:
        top_program_html = BeautifulSoup(f.read(), features="html.parser")

    # Find links to external modules
    uses_box = top_program_html.find(string="Uses").parent.parent.parent
    links = {
        tag.text: tag.a["href"] for tag in uses_box("li", class_="list-inline-item")
    }

    assert len(links) == 3
    assert "external_module" in links
    local_url = urlparse(links["external_module"])
    local_path = program_dir / local_url.path
    assert local_path.is_file()

    assert "remote_module" in links
    remote_url = urlparse(links["remote_module"])
    assert remote_url.scheme == "https"


def test_procedure_module_use_links_(external_project):
    """Check that links to external modules used by functions are correct"""

    top_level_project, _ = external_project

    # Read generated HTML
    proc_dir = top_level_project / "doc/proc"
    with open(proc_dir / "abortcriteria_load.html", "r") as f:
        procedure_html = BeautifulSoup(f.read(), features="html.parser")

    # Find links to external modules
    uses_box = procedure_html.find(string="Uses").parent.parent.parent
    links = {
        tag.text: tag.a["href"] for tag in uses_box("li", class_="list-inline-item")
    }

    assert len(links) == 1
    assert "external_module" in links
    local_url = urlparse(links["external_module"])
    local_path = proc_dir / local_url.path
    assert local_path.is_file()


def test_external_ford_link(external_project):
    """Test #696 -- malformed [[ford]] links to external entities"""

    top_level_project, _ = external_project

    prog_dir = top_level_project / "doc/program"
    with open(prog_dir / "top_level.html") as f:
        prog_html = BeautifulSoup(f.read(), features="html.parser")

    link = prog_html.find("a", string="remote_sub")
    assert link["href"] == "https://example.com/proc/remote_sub.html"
